#!/usr/bin/env python

# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import importlib.metadata
import json
import pathlib
import socket
import traceback
from ast import literal_eval
from base64 import b64decode
from os import environ, path, remove
from pprint import pprint
from typing import Optional

import click
import dotenv
import pandas as pd
import semver
import yaml
from tabulate import tabulate

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas
import mlrun.platforms
import mlrun.utils.helpers
from mlrun.common.helpers import parse_versioned_object_uri
from mlrun.runtimes.mounts import auto_mount as auto_mount_modifier

from .config import config as mlconf
from .db import get_run_db
from .errors import err_to_str
from .model import RunTemplate
from .projects import load_project
from .run import (
    get_object,
    import_function,
    import_function_to_dict,
    load_func_code,
    new_function,
)
from .runtimes import RemoteRuntime, RunError, RuntimeKinds, ServingRuntime
from .secrets import SecretsStore
from .utils import (
    RunKeys,
    dict_to_yaml,
    get_in,
    is_relative_path,
    list2dict,
    logger,
    update_in,
)
from .utils.version import Version

pd.set_option("mode.chained_assignment", None)


def validate_base_argument(ctx: click.Context, param: click.Parameter, value: str):
    # click 8.2 expects the context to be passed to make_metavar
    if semver.VersionInfo.parse(
        importlib.metadata.version("click")
    ) < semver.VersionInfo.parse("8.2.0"):
        metavar_func = functools.partial(param.make_metavar)
    else:
        metavar_func = functools.partial(param.make_metavar, ctx)
    if value and value.startswith("-"):
        raise click.BadParameter(
            f"{param.human_readable_name} ({value}) cannot start with '-', ensure the command options are typed "
            f"correctly. Preferably use '--' to separate options and arguments "
            f"e.g. 'mlrun run --option1 --option2 -- {metavar_func()} [--arg1|arg1] [--arg2|arg2]'",
            ctx=ctx,
            param=param,
        )

    return value


@click.group()
def main():
    pass


@main.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("url", type=str, required=False, callback=validate_base_argument)
@click.option(
    "--param",
    "-p",
    default=[],
    multiple=True,
    help="parameter name and value tuples, e.g. -p x=37 -p y='text'",
)
@click.option("--inputs", "-i", multiple=True, help="input artifact")
@click.option("--outputs", "-o", multiple=True, help="output artifact/result for kfp")
@click.option("--in-path", help="default input path/url (prefix) for artifact")
@click.option("--out-path", help="default output path/url (prefix) for artifact")
@click.option(
    "--secrets",
    "-s",
    multiple=True,
    help="secrets file=<filename> or env=ENV_KEY1,..",
)
@click.option("--uid", help="unique run ID")
@click.option("--name", help="run name")
@click.option(
    "--workflow", help="sets the run labels to match the given workflow name/id"
)
@click.option("--project", help="project name/id")
@click.option("--db", default="", help="save run results to path or DB url")
@click.option(
    "--runtime", "-r", default="", help="function spec dict, for pipeline usage"
)
@click.option(
    "--kfp", is_flag=True, help="running inside Kubeflow Piplines, do not use"
)
@click.option(
    "--hyperparam",
    "-x",
    default=[],
    multiple=True,
    help="hyper parameters (will expand to multiple tasks) e.g. --hyperparam p2=[1,2,3]",
)
@click.option(
    "--param-file", default="", help="path to csv table of execution (hyper) params"
)
@click.option(
    "--selector",
    default="",
    help="how to select the best result from a list, e.g. max.accuracy",
)
@click.option(
    "--hyper-param-strategy",
    default="",
    help="hyperparam tuning strategy list | grid | random",
)
@click.option(
    "--hyper-param-options",
    default="",
    help="hyperparam options json string",
)
@click.option(
    "--func-url",
    "-f",
    default="",
    help="path/url of function yaml or function " "yaml or db://<project>/<name>[:tag]",
)
@click.option("--task", default="", help="path/url to task yaml")
@click.option(
    "--handler", default="", help="invoke function handler inside the code file"
)
@click.option("--mode", help="special run mode ('pass' for using the command as is)")
@click.option("--schedule", help="cron schedule")
@click.option("--from-env", is_flag=True, help="read the spec from the env var")
@click.option("--dump", is_flag=True, help="dump run results as YAML")
@click.option("--image", default="", help="container image (defaults to mlrun/mlrun)")
@click.option("--kind", default="", help="serverless runtime kind")
@click.option("--source", default="", help="source code archive/git")
@click.option("--local", is_flag=True, help="run the task locally (ignore runtime)")
@click.option(
    "--auto-mount", is_flag=True, help="add volume mount to job using auto mount option"
)
@click.option("--workdir", default="", help="run working directory")
@click.option("--origin-file", default="", help="for internal use")
@click.option("--label", multiple=True, help="run labels (key=val)")
@click.option("--watch", "-w", is_flag=True, help="watch/tail run log")
@click.option("--verbose", is_flag=True, help="verbose log")
@click.option(
    "--scrape-metrics",
    is_flag=True,
    help="whether to add the `mlrun/scrape-metrics` label to this run's resources",
)
@click.option(
    "--env-file", default="", help="path to .env file to load config/variables from"
)
@click.option(
    "--auto-build",
    is_flag=True,
    help="when set functions will be built prior to run if needed",
)
@click.argument("run_args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--ensure-project",
    is_flag=True,
    help="ensure the project exists, if not, create project",
)
@click.option(
    "--returns",
    multiple=True,
    help="Logging configurations for the handler's returning values",
)
@click.option(
    "--allow-cross-project",
    is_flag=True,
    default=True,  # TODO: remove this default in 1.11
    help="Override the loaded project name. This flag ensures awareness of loading an existing project yaml "
    "as a baseline for a new project with a different name",
)
def run(
    url,
    param,
    inputs,
    outputs,
    in_path,
    out_path,
    secrets,
    uid,
    name,
    workflow,
    project,
    db,
    runtime,
    kfp,
    hyperparam,
    param_file,
    selector,
    hyper_param_strategy,
    hyper_param_options,
    func_url,
    task,
    handler,
    mode,
    schedule,
    from_env,
    dump,
    image,
    kind,
    source,
    local,
    auto_mount,
    workdir,
    origin_file,
    label,
    watch,
    verbose,
    scrape_metrics,
    env_file,
    auto_build,
    run_args,
    ensure_project,
    returns,
    allow_cross_project,
):
    """Execute a task and inject parameters."""

    if env_file:
        mlrun.set_env_from_file(env_file)

    out_path = out_path or environ.get("MLRUN_ARTIFACT_PATH")
    config = environ.get("MLRUN_EXEC_CONFIG")
    if from_env and config:
        config = json.loads(config)
        # If run is a retry we need to maintain the run status therefore using RunObject instead of RunTemplate
        retry_count = config.get("status", {}).get("retry_count")
        if retry_count:
            logger.info(f"Retrying run - attempt: {retry_count + 1}")
            runobj = mlrun.RunObject.from_dict(config)
        else:
            runobj = RunTemplate.from_dict(config)
    elif task:
        obj = get_object(task)
        task = yaml.load(obj, Loader=yaml.FullLoader)
        runobj = RunTemplate.from_dict(task)
    else:
        runobj = RunTemplate()

    set_item(runobj.metadata, uid, "uid")
    set_item(runobj.metadata, name, "name")
    set_item(runobj.metadata, project, "project")

    if label:
        label_dict = list2dict(label)
        for k, v in label_dict.items():
            runobj.metadata.labels[k] = v

    if workflow:
        runobj.metadata.labels[mlrun_constants.MLRunInternalLabels.workflow] = workflow
        runobj.metadata.labels[mlrun_constants.MLRunInternalLabels.runner_pod] = (
            socket.gethostname()
        )

    if db:
        mlconf.dbpath = db

    # remove potential quotes from command
    eval_url = py_eval(url)
    url = eval_url if isinstance(eval_url, str) else url
    url_file = url
    url_args = ""
    if url:
        split = url.split(maxsplit=1)
        url_file = split[0]
        if len(split) > 1:
            url_args = split[1]

    if ensure_project and project:
        mlrun.get_or_create_project(
            name=project,
            context="./",
            allow_cross_project=allow_cross_project,
        )
    if func_url or kind:
        if func_url:
            runtime = func_url_to_runtime(func_url, ensure_project, allow_cross_project)
            kind = get_in(runtime, "kind", kind or "job")
            if runtime is None:
                exit(1)
        else:
            kind = kind or "job"
            runtime = {"kind": kind, "spec": {"image": image or "mlrun/mlrun"}}

        if kind not in ["", "local", "dask"] and url:
            if url_file and path.isfile(url_file):
                with open(url_file) as fp:
                    body = fp.read()
                based = mlrun.utils.helpers.encode_user_code(body)
                logger.info(f"packing code at {url_file}")
                update_in(runtime, "spec.build.functionSourceCode", based)
                url = f"main{pathlib.Path(url_file).suffix} {url_args}"
                update_in(runtime, "spec.build.code_origin", url_file)
    elif runtime:
        runtime = py_eval(runtime)
        runtime = mlrun.utils.helpers.as_dict(runtime)
        if not isinstance(runtime, dict):
            print(f"Runtime parameter must be a dict, not {type(runtime)}")
            exit(1)
    else:
        runtime = {}

    code = environ.get("MLRUN_EXEC_CODE")
    if get_in(runtime, "kind", "") == "dask":
        code = get_in(runtime, "spec.build.functionSourceCode", code)
    if from_env and code:
        code = b64decode(code).decode("utf-8")
        origin_file = pathlib.Path(
            get_in(runtime, "spec.build.origin_filename", origin_file)
        )
        if kfp:
            print(f"Code:\n{code}\n")
        suffix = pathlib.Path(url_file).suffix if url else ".py"

        # * is a placeholder for the url file when we want to use url args and let mlrun resolve the url file
        if (
            suffix != ".py"
            and mode != "pass"
            and url_file != "{codefile}"
            and url_file != "*"
        ):
            print(
                f"command/url ({url}) must specify a .py file when not in 'pass' mode"
            )
            exit(1)
        if mode == "pass":
            if "{codefile}" in url:
                url_file = origin_file.name or "codefile"
                url = url.replace("{codefile}", url_file)
            elif suffix == ".sh" or origin_file.suffix == ".sh":
                url_file = origin_file.name or "codefile.sh"
                url = f"bash {url_file} {url_args}".strip()
            else:
                print(
                    "Error: command must be specified with '{codefile}' in it "
                    "(to determine the position of the code file)"
                )
                exit(1)
        else:
            url_file = "main.py"
            if origin_file.name:
                url_file = origin_file.stem + ".py"
            url = f"{url_file} {url_args}".strip()
        with open(url_file, "w") as fp:
            fp.write(code)

    # at this point the url placeholder should have been resolved to the actual url file
    if url == "*":
        print("command/url '*' placeholder is not allowed when code is not from env")
        exit(1)

    if url:
        if not name and not runtime:
            name = path.splitext(path.basename(url))[0]
            runobj.metadata.name = runobj.metadata.name or name
        update_in(runtime, "spec.command", url)

    if run_args:
        update_in(runtime, "spec.args", list(run_args))

    update_in(runtime, "spec.image", image or "mlrun/mlrun", replace=bool(image))

    set_item(runobj.spec, handler, "handler")
    set_item(runobj.spec, param, "parameters", fill_params(param))

    set_item(runobj.spec, hyperparam, "hyperparams", fill_params(hyperparam))
    if hyper_param_options:
        runobj.spec.hyper_param_options = py_eval(hyper_param_options)
    set_item(runobj.spec.hyper_param_options, param_file, "param_file")
    set_item(runobj.spec.hyper_param_options, hyper_param_strategy, "strategy")
    set_item(runobj.spec.hyper_param_options, selector, "selector")

    set_item(runobj.spec, inputs, RunKeys.inputs, list2dict(inputs))
    set_item(
        runobj.spec, returns, RunKeys.returns, [py_eval(value) for value in returns]
    )
    set_item(runobj.spec, in_path, RunKeys.input_path)
    set_item(runobj.spec, out_path, RunKeys.output_path)
    set_item(runobj.spec, outputs, RunKeys.outputs, list(outputs))
    set_item(
        runobj.spec, secrets, RunKeys.secrets, line2keylist(secrets, "kind", "source")
    )
    set_item(runobj.spec, verbose, "verbose")
    set_item(runobj.spec, scrape_metrics, "scrape_metrics")
    update_in(runtime, "metadata.name", name, replace=False)
    update_in(runtime, "metadata.project", project, replace=False)
    if not kind and "." in handler:
        # handle the case of module.submodule.handler
        update_in(runtime, "kind", "local")

    if kfp or runobj.spec.verbose or verbose:
        print(f"MLRun version: {str(Version().get())}")
        print("Runtime:")
        pprint(runtime)
        print("Run:")
        pprint(runobj.to_dict())

    try:
        fn = new_function(runtime=runtime, kfp=kfp, mode=mode, source=source)
        if workdir:
            fn.spec.workdir = workdir
        if auto_mount:
            fn.apply(auto_mount_modifier())
        fn.is_child = from_env and not kfp
        if kfp:
            # if pod is running inside kfp pod, we don't really need the run logs to be printed actively, we can just
            # pull the run state, and pull the logs periodically
            # we will set watch=None only when the pod is running inside kfp, and this tells the run to pull state
            # and logs periodically
            # TODO: change watch to be a flag with more options (with_logs, wait_for_completion, etc.)
            watch = watch or None
        resp = fn.run(
            runobj,
            watch=watch,
            schedule=schedule,
            local=local,
            auto_build=auto_build,
            project=project,
        )
        if resp and dump:
            print(resp.to_yaml())
    except RunError as err:
        print(f"Runtime error: {err_to_str(err)}")
        exit(1)


@main.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("func_url", type=str, required=False, callback=validate_base_argument)
@click.option("--name", help="function name")
@click.option("--project", help="project name")
@click.option("--tag", default="", help="function tag")
@click.option("--image", "-i", help="target image path")
@click.option(
    "--source", "-s", default="", help="location/url of the source files dir/tar"
)
@click.option("--base-image", "-b", help="base docker image")
@click.option(
    "--command",
    "-c",
    default=[],
    multiple=True,
    help="build commands, e.g. '-c pip install pandas'",
)
@click.option("--secret-name", default="", help="container registry secret name")
@click.option("--archive", "-a", default="", help="destination archive for code (tar)")
@click.option("--silent", is_flag=True, help="do not show build logs")
@click.option("--with-mlrun", is_flag=True, help="add MLRun package")
@click.option("--db", default="", help="save run results to DB url")
@click.option(
    "--runtime", "-r", default="", help="function spec dict, for pipeline usage"
)
@click.option(
    "--kfp", is_flag=True, help="running inside Kubeflow Pipelines, do not use"
)
@click.option("--skip", is_flag=True, help="skip if already deployed")
@click.option(
    "--env-file", default="", help="path to .env file to load config/variables from"
)
@click.option(
    "--ensure-project",
    is_flag=True,
    help="ensure the project exists, if not, create project",
)
@click.option(
    "--state-file-path", default="/tmp/state", help="path to file with state data"
)
@click.option(
    "--image-file-path", default="/tmp/image", help="path to file with image data"
)
@click.option(
    "--full-image-file-path",
    default="/tmp/fullimage",
    help="path to file with full image data",
)
@click.option(
    "--allow-cross-project",
    is_flag=True,
    default=True,  # TODO: remove this default in 1.11
    help="Override the loaded project name. This flag ensures awareness of loading an existing project yaml "
    "as a baseline for a new project with a different name",
)
def build(
    func_url,
    name,
    project,
    tag,
    image,
    source,
    base_image,
    command,
    secret_name,
    archive,
    silent,
    with_mlrun,
    db,
    runtime,
    kfp,
    skip,
    env_file,
    ensure_project,
    state_file_path,
    image_file_path,
    full_image_file_path,
    allow_cross_project,
):
    """Build a container image from code and requirements."""

    if env_file:
        mlrun.set_env_from_file(env_file)

    if db:
        mlconf.dbpath = db

    if runtime:
        runtime = py_eval(runtime)
        runtime = mlrun.utils.helpers.as_dict(runtime)
        if not isinstance(runtime, dict):
            print(f"Runtime parameter must be a dict, not {type(runtime)}")
            exit(1)
        if kfp:
            print("Runtime:")
            pprint(runtime)
        # use kind = "job" by default if not specified
        runtime.setdefault("kind", "job")
        func = new_function(runtime=runtime)

    elif func_url:
        if func_url.startswith("db://"):
            func_url = func_url[5:]
        elif func_url == ".":
            func_url = "function.yaml"
        func = import_function(func_url)

    else:
        print("Error: Function path or url are required")
        exit(1)

    meta = func.metadata
    meta.project = project or meta.project or mlconf.active_project
    meta.name = name or meta.name
    meta.tag = tag or meta.tag

    b = func.spec.build
    if func.kind not in ["", "local"]:
        b.base_image = base_image or b.base_image
        b.commands = list(command) or b.commands
        b.image = image or b.image
        b.secret = secret_name or b.secret

    if source.endswith(".py"):
        if not path.isfile(source):
            print(f"Source file doesnt exist ({source})")
            exit(1)
        with open(source) as fp:
            body = fp.read()
        based = mlrun.utils.helpers.encode_user_code(body)
        logger.info(f"Packing code at {source}")
        b.functionSourceCode = based
        func.spec.command = ""
    else:
        b.source = source or b.source
        # todo: upload stuff

    archive = archive or mlconf.default_archive
    if archive:
        src = b.source or "./"
        logger.info(f"Uploading data from {src} to {archive}")
        target = archive if archive.endswith("/") else archive + "/"
        target += f"src-{meta.project}-{meta.name}-{meta.tag or 'latest'}.tar.gz"
        mlrun.datastore.utils.upload_tarball(src, target)
        # todo: replace function.yaml inside the tar
        b.source = target

    with_mlrun = True if with_mlrun else None  # False will map to None

    if ensure_project and project:
        mlrun.get_or_create_project(
            name=project,
            context="./",
            allow_cross_project=allow_cross_project,
        )

    if hasattr(func, "deploy"):
        logger.info("Remote deployment started")
        try:
            func.deploy(
                with_mlrun=with_mlrun, watch=not silent, is_kfp=kfp, skip_deployed=skip
            )
        except Exception as err:
            print(f"Deploy error, {err_to_str(err)}")
            exit(1)

        state = func.status.state
        image = func.spec.image
        if kfp:
            with open(state_file_path, "w") as fp:
                fp.write(state or "none")
            full_image = func.full_image_path(image) or ""
            with open(image_file_path, "w") as fp:
                fp.write(image)
            with open(full_image_file_path, "w") as fp:
                fp.write(full_image)
            print("Full image path = ", full_image)

        print(f"Function built, state={state} image={image}")
    else:
        print("Function does not have a deploy() method")
        exit(1)


@main.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("spec", type=str, required=False, callback=validate_base_argument)
@click.option("--source", "-s", default="", help="location/url of the source")
@click.option(
    "--func-url",
    "-f",
    default="",
    help="path/url of function yaml or function " "yaml or db://<project>/<name>[:tag]",
)
@click.option("--project", "-p", default="", help="project name")
@click.option("--model", "-m", multiple=True, help="model name and path (name=path)")
@click.option("--kind", "-k", default=None, help="runtime sub kind")
@click.option("--tag", default="", help="version tag")
@click.option("--env", "-e", multiple=True, help="environment variables")
@click.option("--verbose", is_flag=True, help="verbose log")
@click.option(
    "--env-file", default="", help="path to .env file to load config/variables from"
)
@click.option(
    "--ensure-project",
    is_flag=True,
    help="ensure the project exists, if not, create project",
)
@click.option(
    "--allow-cross-project",
    is_flag=True,
    default=True,  # TODO: remove this default in 1.11
    help="Override the loaded project name. This flag ensures awareness of loading an existing project yaml "
    "as a baseline for a new project with a different name",
)
def deploy(
    spec,
    source,
    func_url,
    project,
    model,
    tag,
    kind,
    env,
    verbose,
    env_file,
    ensure_project,
    allow_cross_project,
):
    """Deploy model or function"""
    if env_file:
        mlrun.set_env_from_file(env_file)

    if ensure_project and project:
        mlrun.get_or_create_project(
            name=project,
            context="./",
            allow_cross_project=allow_cross_project,
        )

    if func_url:
        runtime = func_url_to_runtime(func_url, ensure_project, allow_cross_project)
        if runtime is None:
            exit(1)
    elif spec:
        runtime = py_eval(spec)
    else:
        runtime = {}

    runtime = mlrun.utils.helpers.as_dict(runtime)
    if not isinstance(runtime, dict):
        print(f"Runtime parameter must be a dict, not {type(runtime)}")
        exit(1)

    if verbose:
        pprint(runtime)
        pprint(model)

    # support both v1 & v2+ model struct for backwards compatibility
    if runtime and runtime["kind"] == RuntimeKinds.serving:
        print("Deploying V2 model server")
        function = ServingRuntime.from_dict(runtime)
        if model:
            # v2+ model struct (list of json obj)
            for _model in model:
                args = json.loads(_model)
                function.add_model(**args)
    else:
        function = RemoteRuntime.from_dict(runtime)
        if kind:
            function.spec.function_kind = kind
        if model:
            # v1 model struct (list of k=v)
            models = list2dict(model)
            for k, v in models.items():
                function.add_model(k, v)

    function.spec.source = source
    if env:
        for k, v in list2dict(env).items():
            function.set_env(k, v)
    function.verbose = verbose

    try:
        addr = function.deploy(project=project, tag=tag)
    except Exception as err:
        print(f"deploy error: {err_to_str(err)}")
        exit(1)

    print(f"Function deployed, address={addr}")
    with open("/tmp/output", "w") as fp:
        fp.write(addr)
    with open("/tmp/name", "w") as fp:
        fp.write(function.status.nuclio_name)


@main.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("kind", type=str, callback=validate_base_argument)
@click.argument(
    "name",
    type=str,
    default="",
    required=False,
    callback=validate_base_argument,
)
@click.option("--selector", "-s", default="", help="label selector")
@click.option("--namespace", "-n", help="kubernetes namespace")
@click.option("--uid", help="unique ID")
@click.option("--project", "-p", help="project name")
@click.option("--tag", "-t", default="", help="artifact/function tag")
@click.option("--db", help="db path/url")
@click.argument("extra_args", nargs=-1, type=click.UNPROCESSED)
def get(kind, name, selector, namespace, uid, project, tag, db, extra_args):
    """List/get one or more object per kind/class.

    KIND - resource type to list/get: run | runtime | workflow | artifact | function
    NAME - optional, resource name or category
    """

    if db:
        mlconf.dbpath = db
    if not project:
        logger.warning(
            "Project parameter was not specified. Defaulting to 'default' project"
        )
    if kind.startswith("po"):
        logger.warning("Unsupported, use 'get runtimes' instead")
        return

    elif kind.startswith("runtime"):
        run_db = get_run_db(db or mlconf.dbpath)
        # the name field is used as function kind, set to None if empty
        name = name if name else None
        runtimes = run_db.list_runtime_resources(
            label_selector=selector, kind=name, project=project
        )
        for runtime in runtimes:
            print(dict_to_yaml(runtime.dict()))
            print()

    elif kind.startswith("run"):
        if tag:
            print(
                "Unsupported argument '--tag' for listing runs. Perhaps you should use '--selector' instead"
            )
            return

        run_db = get_run_db()
        if name:
            run = run_db.read_run(name, project=project)
            print(dict_to_yaml(run))
            return

        runs = run_db.list_runs(uid=uid, project=project, labels=selector)
        df = runs.to_df()[
            ["name", "uid", "iter", "start", "end", "state", "parameters", "results"]
        ]
        # df['uid'] = df['uid'].apply(lambda x: f'..{x[-6:]}')
        for time_column in ["start", "end"]:
            df[time_column] = df[time_column].apply(time_str)
        df["parameters"] = df["parameters"].apply(dict_to_str)
        df["results"] = df["results"].apply(dict_to_str)
        print(tabulate(df, headers="keys"))

    elif kind.startswith("art"):
        run_db = get_run_db()
        artifacts = run_db.list_artifacts(
            name, project=project, tag=tag, labels=selector
        )
        df = artifacts.to_df()[
            ["key", "iter", "kind", "path", "hash", "updated", "uri", "tree"]
        ]
        df["tree"] = df["tree"].apply(lambda x: f"..{x[-8:]}")
        df["hash"] = df["hash"].apply(lambda x: f"..{x[-6:]}")
        df["updated"] = df["updated"].apply(time_str)
        df.rename(columns={"tree": "job/workflow uid"}, inplace=True)
        print(tabulate(df, headers="keys"))

    elif kind.startswith("func"):
        run_db = get_run_db()
        if name:
            f = run_db.get_function(name, project=project, tag=tag)
            print(dict_to_yaml(f))
            return

        functions = run_db.list_functions(name, project=project, labels=selector)
        lines = []
        headers = ["kind", "state", "name:tag", "hash"]
        for f in functions:
            name = get_in(f, "metadata.name")
            tag = get_in(f, "metadata.tag", "")
            line = [
                get_in(f, "kind", ""),
                get_in(f, "status.state", ""),
                f"{name}:{tag}",
                get_in(f, "metadata.hash", ""),
            ]
            lines.append(line)
        print(tabulate(lines, headers=headers))

    elif kind.startswith("workflow"):
        run_db = get_run_db()
        if project == "*":
            print("Warning, reading workflows for all projects may take a long time !")
            pipelines = run_db.list_pipelines(project=project, page_size=200)
            pipe_runs = pipelines.runs
            while pipelines.next_page_token is not None:
                pipelines = run_db.list_pipelines(
                    project=project, page_token=pipelines.next_page_token
                )
                pipe_runs.extend(pipelines.runs)
        else:
            pipelines = run_db.list_pipelines(project=project)
            pipe_runs = pipelines.runs

        lines = []
        headers = ["project", "name", "status", "created at", "finished at"]
        for pipe_run in pipe_runs:
            line = [
                pipe_run["project"],
                pipe_run["name"],
                pipe_run["status"],
                pipe_run["created_at"],
                pipe_run["finished_at"],
            ]
            lines.append(line)
        print(tabulate(lines, headers=headers))

    else:
        print(
            "Currently only get runs | runtimes | workflows | artifacts  | func [name] | runtime are supported"
        )


@main.command()
def version():
    """get mlrun version"""
    print(f"MLRun version: {str(Version().get())}")


@main.command()
@click.argument("uid", type=str)
@click.option(
    "--project", "-p", help="project name (defaults to mlrun.mlconf.active_project)"
)
@click.option("--offset", type=int, default=0, help="byte offset")
@click.option("--db", help="api and db service path/url")
def logs(uid, project, offset, db):
    """Get or watch task logs"""
    mldb = get_run_db(db or mlconf.dbpath)
    if mldb.kind == "http":
        state, _ = mldb.watch_log(uid, project, watch=False, offset=offset)
    else:
        state, text = mldb.get_log(uid, project, offset=offset)
        if text:
            print(text.decode())

    if state:
        print(f"Final state: {state}")


@main.command()
@click.argument("context", default="", type=str, required=False)
@click.option("--name", "-n", help="project name")
@click.option("--url", "-u", help="remote git or archive url")
@click.option("--run", "-r", help="run workflow name of .py file")
@click.option(
    "--arguments",
    "-a",
    default=[],
    multiple=True,
    help="Kubeflow pipeline arguments name and value tuples (with -r flag), e.g. -a x=6",
)
@click.option("--artifact-path", "-p", help="output artifacts path")
@click.option(
    "--param",
    "-x",
    default=[],
    multiple=True,
    help="mlrun project parameter name and value tuples, e.g. -p x=37 -p y='text'",
)
@click.option(
    "--secrets",
    "-s",
    multiple=True,
    help="secrets file=<filename> or env=ENV_KEY1,..",
)
@click.option("--namespace", help="k8s namespace")
@click.option("--db", help="api and db service path/url")
@click.option("--init-git", is_flag=True, help="for new projects init git context")
@click.option(
    "--clone", "-c", is_flag=True, help="force override/clone into the context dir"
)
@click.option("--sync", is_flag=True, help="sync functions into db")
@click.option(
    "--watch", "-w", is_flag=True, help="wait for pipeline completion (with -r flag)"
)
@click.option(
    "--dirty", "-d", is_flag=True, help="allow run with uncommitted git changes"
)
@click.option("--git-repo", help="git repo (org/repo) for git comments")
@click.option(
    "--git-issue", type=int, default=None, help="git issue number for git comments"
)
@click.option("--handler", default=None, help="workflow function handler name")
@click.option("--engine", default=None, help="workflow engine (kfp/local)")
@click.option("--local", is_flag=True, help="try to run workflow functions locally")
@click.option(
    "--timeout",
    type=int,
    default=None,
    help="timeout in seconds to wait for pipeline completion (used when watch=True)",
)
@click.option(
    "--env-file", default="", help="path to .env file to load config/variables from"
)
@click.option(
    "--save/--no-save",
    default=True,
    help="create and save the project if not exist",
)
@click.option(
    "--schedule",
    type=str,
    default=None,
    help="To create a schedule define a standard crontab expression string."
    "for help see: "
    "https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html#module-apscheduler.triggers.cron."
    "For using the pre-defined workflow's schedule, set --schedule 'true'",
)
@click.option(
    "--save-secrets",
    is_flag=True,
    help="Store the project secrets as k8s secrets",
)
@click.option(
    "--notifications",
    "--notification",
    "-nt",
    multiple=True,
    help="To have a notification for the run set notification file "
    "destination define: file=notification.json or a "
    'dictionary configuration e.g \'{"slack":{"webhook":"<webhook>"}}\'',
)
@click.option(
    "--allow-cross-project",
    is_flag=True,
    default=True,  # TODO: remove this default in 1.11
    help="Override the loaded project name. This flag ensures awareness of loading an existing project yaml "
    "as a baseline for a new project with a different name",
)
def project(
    context,
    name,
    url,
    run,
    arguments,
    artifact_path,
    param,
    secrets,
    namespace,
    db,
    init_git,
    clone,
    sync,
    watch,
    dirty,
    git_repo,
    git_issue,
    handler,
    engine,
    local,
    env_file,
    timeout,
    schedule,
    notifications,
    save_secrets,
    save,
    allow_cross_project,
):
    """load and/or run a project"""
    if env_file:
        mlrun.set_env_from_file(env_file)

    if db:
        mlconf.dbpath = db

    # set the CLI/GIT parameters in load_project() so they can be used by project setup scripts
    parameters = fill_params(param) if param else {}
    if git_repo:
        parameters["git_repo"] = git_repo
    if git_issue:
        parameters["git_issue"] = git_issue
    commit = environ.get("GITHUB_SHA") or environ.get("CI_COMMIT_SHA")
    if commit and not parameters.get("commit_id"):
        parameters["commit_id"] = commit

    proj = load_project(
        context,
        url,
        name,
        init_git=init_git,
        clone=clone,
        save=save,
        parameters=parameters,
        allow_cross_project=allow_cross_project,
    )
    url_str = " from " + url if url else ""
    print(f"Loading project {proj.name}{url_str} into {context}:\n")

    if is_relative_path(artifact_path):
        artifact_path = path.abspath(artifact_path)
    if secrets:
        secrets = line2keylist(secrets, "kind", "source")
        secret_store = SecretsStore.from_list(secrets)
        # Used to run a workflow with secrets in runtime, without using or storing k8s secrets.
        # To run a scheduled workflow or to use those secrets in other runs, save
        # the secrets in k8s and use the --save-secret flag
        proj._secrets = secret_store
        if save_secrets:
            proj.set_secrets(secret_store._secrets)
    print(proj.to_yaml())

    if run:
        if schedule is not None and schedule.lower() in ["1", "yes", "y", "t", "true"]:
            schedule = True
        workflow_path = None
        if run.endswith(".py"):
            workflow_path = run
            run = None

        args = None
        if arguments:
            args = fill_params(arguments)

        print(f"Running workflow {run} file: {workflow_path}")
        gitops = (
            git_issue
            or environ.get("GITHUB_EVENT_PATH")
            or environ.get("CI_MERGE_REQUEST_IID")
        )
        if gitops:
            proj.notifiers.add_notification(
                "git",
                {
                    "repo": git_repo,
                    "issue": git_issue,
                    "token": proj.get_param("GIT_TOKEN"),
                },
            )
        if notifications:
            load_notification(notifications, proj)
        try:
            proj.run(
                name=run,
                workflow_path=workflow_path,
                arguments=args,
                artifact_path=artifact_path,
                namespace=namespace,
                sync=sync,
                watch=watch,
                dirty=dirty,
                workflow_handler=handler,
                engine=engine,
                local=local,
                schedule=schedule,
                timeout=timeout,
            )
        except Exception as err:
            print(traceback.format_exc())
            send_workflow_error_notification(run, proj, err)
            exit(1)

    elif sync:
        print("Saving project functions to db ..")
        proj.sync_functions(save=True)


def validate_runtime_kind(ctx, param, value):
    possible_kinds = RuntimeKinds.runtime_with_handlers()
    if value is not None and value not in possible_kinds:
        raise click.BadParameter(
            f"kind must be one of {possible_kinds}", ctx=ctx, param=param
        )
    return value


@main.command()
@click.argument("kind", callback=validate_runtime_kind, default=None, required=False)
@click.argument("object_id", metavar="id", type=str, default=None, required=False)
@click.option("--api", help="api service url")
@click.option("--label-selector", "-ls", default="", help="label selector")
@click.option(
    "--force", "-f", is_flag=True, help="clean resources in non-terminal states as well"
)
@click.option(
    "--grace-period",
    "-gp",
    type=int,
    # When someone triggers the cleanup manually we assume they want runtime resources in terminal state to be removed
    # now, therefore not using here mlconf.runtime_resources_deletion_grace_period
    default=0,
    help="the grace period (in seconds) that will be given to runtime resources (after they're in terminal state) "
    "before cleaning them. Ignored when --force is given",
    show_default=True,
)
def clean(kind, object_id, api, label_selector, force, grace_period):
    """
    Clean jobs resources

    \b
    Examples:

        \b
        # Clean resources for all runs of all runtimes
        mlrun clean

        \b
        # Clean resources for all runs of a specific kind (e.g. job)
        mlrun clean job

        \b
        # Clean resources for specific job (by uid)
        mlrun clean mpijob 15d04c19c2194c0a8efb26ea3017254b
    """
    mldb = get_run_db(api or mlconf.dbpath)
    mldb.delete_runtime_resources(
        kind=kind,
        object_id=object_id,
        label_selector=label_selector,
        force=force,
        grace_period=grace_period,
    )


@main.command(name="watch-stream")
@click.argument("url", type=str)
@click.option(
    "--shard-ids",
    "-s",
    multiple=True,
    type=int,
    help="shard id to listen on (can be multiple)",
)
@click.option("--seek", help="where to start/seek (EARLIEST or LATEST)")
@click.option(
    "--interval",
    "-i",
    default=3,
    show_default=True,
    help="interval in seconds",
    type=int,
)
@click.option(
    "--is-json",
    "-j",
    is_flag=True,
    help="indicate the payload is json (will be deserialized)",
)
def watch_stream(url, shard_ids, seek, interval, is_json):
    """watch on a stream and print data every interval"""
    mlrun.platforms.watch_stream(
        url, shard_ids, seek, interval=interval, is_json=is_json
    )


@main.command(name="config")
@click.argument("command", type=str, default="", required=False)
@click.option(
    "--env-file",
    "-f",
    default="",
    help="path to the mlrun .env file (defaults to '~/.mlrun.env')",
)
@click.option("--api", "-a", type=str, help="api service url")
@click.option("--artifact-path", "-p", help="default artifacts path")
@click.option("--username", "-u", help="username (for remote access)")
@click.option("--access-key", "-k", help="access key (for remote access)")
@click.option(
    "--env-vars",
    "-e",
    default=[],
    multiple=True,
    help="additional env vars, e.g. -e AWS_ACCESS_KEY_ID=<key-id>",
)
def show_or_set_config(
    command, env_file, api, artifact_path, username, access_key, env_vars
):
    """get or set mlrun config

    \b
    Commands:
        get (default) - list the local or remote configuration
                        (can specify the remote api + credentials or an env_file)
        set           - set configuration parameters in mlrun default or specified .env file
        clear         - delete the default or specified config .env file

    \b
    Examples:
        # read the default config
        mlrun config
        # read config using an env file (with connection details)
        mlrun config -f mymlrun.env
        # set configuration and write it to the default env file (~/.mlrun.env)
        mlrun config set -a http://localhost:8080 -u joe -k mykey -e AWS_ACCESS_KEY_ID=<key-id>

    """
    op = command.lower()
    if not op or op == "get":
        # print out the configuration (default or based on the specified env/api)
        if env_file and not path.isfile(path.expanduser(env_file)):
            print(f"Error: Env file {env_file} does not exist")
            exit(1)
        if env_file or api:
            mlrun.set_environment(
                api,
                artifact_path=artifact_path,
                access_key=access_key,
                username=username,
                env_file=env_file,
            )
        print(mlconf.dump_yaml())

    elif op == "set":
        # update the env settings in the default or specified .env file
        filename = path.expanduser(env_file or mlrun.config.default_env_file)
        if not path.isfile(filename):
            print(
                f".env file {filename} not found, creating new and setting configuration"
            )
        else:
            print(f"Updating configuration in .env file {filename}")
        env_dict = {
            "MLRUN_DBPATH": api,
            "MLRUN_ARTIFACT_PATH": artifact_path,
            "V3IO_USERNAME": username,
            "V3IO_ACCESS_KEY": access_key,
        }
        for key, value in env_dict.items():
            if value:
                dotenv.set_key(filename, key, value, quote_mode="always")
        if env_vars:
            for key, value in list2dict(env_vars).items():
                dotenv.set_key(filename, key, value, quote_mode="always")
        if env_file:
            # if its not the default file print the usage details
            print(
                f"To use the {env_file} .env file add the following to your development environment:\n"
                f"MLRUN_ENV_FILE={env_file}"
            )

    elif op == "clear":
        filename = path.expanduser(env_file or mlrun.config.default_env_file)
        if not path.isfile(filename):
            print(f".env file {filename} not found")
        else:
            print(f"Deleting .env file {filename}")
            remove(filename)

    else:
        print(f"Error: Unsupported config option {op}")


def fill_params(params, params_dict=None):
    params_dict = params_dict or {}
    for param in params:
        i = param.find("=")
        if i == -1:
            continue
        key, value = param[:i].strip(), param[i + 1 :].strip()
        if key is None:
            raise ValueError(f"cannot find param key in line ({param})")
        params_dict[key] = py_eval(value)
    return params_dict


def py_eval(data):
    try:
        value = literal_eval(data)
        return value
    except (SyntaxError, ValueError):
        return data


def set_item(obj, item, key, value=None):
    if item:
        if value:
            setattr(obj, key, value)
        else:
            setattr(obj, key, item)


def line2keylist(lines: list, keyname="key", valname="path"):
    out = []
    for line in lines:
        i = line.find("=")
        if i == -1:
            raise ValueError(f'cannot find "=" in line ({keyname}={valname})')
        key, value = line[:i].strip(), line[i + 1 :].strip()
        if key is None:
            raise ValueError(f"cannot find key in line ({keyname}={valname})")
        value = path.expandvars(value)
        out += [{keyname: key, valname: value}]
    return out


def time_str(x):
    try:
        return x.strftime("%b %d %H:%M:%S")
    except ValueError:
        return ""


def dict_to_str(struct: dict):
    if not struct:
        return []
    return ",".join([f"{k}={v}" for k, v in struct.items()])


def func_url_to_runtime(
    func_url,
    ensure_project: bool = False,
    allow_cross_project: Optional[bool] = None,
):
    try:
        if func_url.startswith("db://"):
            func_url = func_url[5:]
            project_instance, name, tag, hash_key = parse_versioned_object_uri(func_url)
            run_db = get_run_db(mlconf.dbpath)
            runtime = run_db.get_function(name, project_instance, tag, hash_key)
        elif func_url == "." or func_url.endswith(".yaml"):
            func_url = "function.yaml" if func_url == "." else func_url
            runtime = import_function_to_dict(func_url, {})
        else:
            mlrun_project = load_project(
                ".", save=ensure_project, allow_cross_project=allow_cross_project
            )
            function = mlrun_project.get_function(func_url, enrich=True)
            if function.kind == "local":
                command, function = load_func_code(function)
                function.spec.command = command
            runtime = function.to_dict()
    except Exception as exc:
        logger.error(f"function {func_url} not found, {err_to_str(exc)}")
        return None

    if not runtime:
        logger.error(f"function {func_url} not found or is null")
        return None

    return runtime


def load_notification(notifications: str, project: mlrun.projects.MlrunProject):
    """
    A dictionary or json file containing notification dictionaries can be used by the user to set notifications.
    Each notification is stored in a tuple called notifications.
    The code then goes through each value in the notifications tuple and check
    if the notification starts with "file=", such as "file=notification.json," in those cases it loads the
    notification.json file and uses add_notification_to_project to add the notifications from the file to
    the project. If not, it adds the notification dictionary to the project.
    :param notifications:  Notifications file or a dictionary to be added to the project
    :param project: The object to which the notifications will be added
    :return:
    """
    for notification in notifications:
        if notification.startswith("file="):
            file_path = notification.split("=")[-1]
            notification = open(file_path)
            notification = json.load(notification)
        else:
            notification = json.loads(notification)
        add_notification_to_project(notification, project)


def add_notification_to_project(
    notification: str, project: mlrun.projects.MlrunProject
):
    for notification_type, notification_params in notification.items():
        project.notifiers.add_notification(
            notification_type=notification_type, params=notification_params
        )


def send_workflow_error_notification(
    run_id: str, mlproject: mlrun.projects.MlrunProject, error: Exception
):
    message = (
        f":x: Failed to run scheduled workflow {run_id} in Project {mlproject.name} !\n"
        f"error: ```{err_to_str(error)}```"
    )
    mlproject.notifiers.push(
        message=message, severity=mlrun.common.schemas.NotificationSeverity.ERROR
    )


if __name__ == "__main__":
    main()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import Optional, Union

import pandas as pd
from storey import EmitEveryEvent, EmitPolicy

import mlrun
import mlrun.common.schemas

from ..config import config as mlconf
from ..data_types import InferOptions
from ..datastore import get_store_uri
from ..datastore.sources import BaseSourceDriver, source_kind_to_driver
from ..datastore.targets import (
    TargetTypes,
    get_default_targets,
    get_offline_target,
    get_online_target,
    get_target_driver,
    update_targets_run_id_for_ingest,
    validate_target_list,
    validate_target_placement,
)
from ..features import Entity, Feature
from ..model import (
    DataSource,
    DataTarget,
    DataTargetBase,
    ModelObj,
    ObjectDict,
    ObjectList,
    VersionedObjMetadata,
)
from ..runtimes import BaseRuntime
from ..runtimes.function_reference import FunctionReference
from ..serving.states import BaseStep, RootFlowStep, previous_step, queue_class_names
from ..serving.utils import StepToDict
from ..utils import StorePrefix, logger
from .common import RunConfig, verify_feature_set_permissions

aggregates_step = "Aggregates"


class FeatureAggregation(ModelObj):
    """feature aggregation requirements"""

    def __init__(
        self, name=None, column=None, operations=None, windows=None, period=None
    ):
        self.name = name
        self.column = column
        self.operations = operations or []
        self.windows = windows or []
        self.period = period


class FeatureSetSpec(ModelObj):
    def __init__(
        self,
        owner=None,
        description=None,
        entities=None,
        features=None,
        partition_keys=None,
        timestamp_key=None,
        label_column=None,
        relations=None,
        source=None,
        targets=None,
        graph=None,
        function=None,
        analysis=None,
        engine=None,
        output_path=None,
        passthrough=None,
    ):
        """Feature set spec object, defines the feature-set's configuration.

        .. warning::
            This class should not be modified directly. It is managed by the parent feature-set object or using
            feature-store APIs. Modifying the spec manually may result in unpredictable behaviour.

        :param description:   text description (copied from parent feature-set)
        :param entities:      list of entity (index key) names or :py:class:`~mlrun.features.FeatureSet.Entity`
        :param features: list of features - :py:class:`~mlrun.features.FeatureSet.Feature`
        :param partition_keys: list of fields to partition results by (other than the default timestamp key)
        :param timestamp_key: timestamp column name
        :param label_column: name of the label column (the one holding the target (y) values)
        :param targets: list of data targets
        :param graph: the processing graph
        :param function: MLRun runtime to execute the feature-set in
        :param engine: name of the processing engine (storey, pandas, or spark), defaults to storey
        :param output_path: default location where to store results (defaults to MLRun's artifact path)
        :param passthrough: if true, ingest will skip offline targets, and get_offline_features will
               read directly from source
        """
        self._features: ObjectList = None
        self._entities: ObjectList = None
        self._targets: ObjectList = None
        self._graph: RootFlowStep = None
        self._source = None
        self._engine = None
        self._function: FunctionReference = None
        self._relations: ObjectDict = None

        self.owner = owner
        self.description = description
        self.entities: list[Union[Entity, str]] = entities or []
        self.relations: dict[str, Union[Entity, str]] = relations or {}
        self.features: list[Feature] = features or []
        self.partition_keys = partition_keys or []
        self.timestamp_key = timestamp_key
        self.source = source
        self.targets = targets or []
        self.graph = graph
        self.label_column = label_column
        self.function = function
        self.analysis = analysis or {}
        self.engine = engine
        self.output_path = output_path or mlconf.artifact_path
        self.passthrough = passthrough
        self.with_default_targets = True

    @property
    def entities(self) -> list[Entity]:
        """feature set entities (indexes)"""
        return self._entities

    @entities.setter
    def entities(self, entities: list[Union[Entity, str]]):
        if entities:
            # if the entity is a string, convert it to Entity class
            for i, entity in enumerate(entities):
                if isinstance(entity, str):
                    entities[i] = Entity(entity)
                elif isinstance(entity, Entity) and entity.name is None:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "You have to provide an "
                        "Entity with valid name of string type"
                    )
                elif isinstance(entity, dict) and (
                    "name" not in entity
                    or ("name" in entity and entity["name"] is None)
                ):
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "You have to provide an "
                        "Entity with valid name of string type"
                    )
        self._entities = ObjectList.from_list(Entity, entities)

    @property
    def features(self) -> list[Feature]:
        """feature set features list"""
        return self._features

    @features.setter
    def features(self, features: list[Feature]):
        self._features = ObjectList.from_list(Feature, features)

    @property
    def targets(self) -> list[DataTargetBase]:
        """list of desired targets (material storage)"""
        return self._targets

    @targets.setter
    def targets(self, targets: list[DataTargetBase]):
        self._targets = ObjectList.from_list(DataTargetBase, targets)

    @property
    def engine(self) -> str:
        """feature set processing engine (storey, pandas, spark)"""
        return self._engine

    @engine.setter
    def engine(self, engine: str):
        engine_list = ["pandas", "spark", "storey"]
        engine = engine if engine else "storey"
        if engine not in engine_list:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"engine must be one of {','.join(engine_list)}"
            )
        self.graph.engine = "sync" if engine and engine in ["pandas", "spark"] else None
        self._engine = engine

    @property
    def graph(self) -> RootFlowStep:
        """feature set transformation graph/DAG"""
        return self._graph

    @graph.setter
    def graph(self, graph):
        self._graph = self._verify_dict(graph, "graph", RootFlowStep)
        self._graph.engine = (
            "sync" if self.engine and self.engine in ["pandas", "spark"] else None
        )

    @property
    def function(self) -> FunctionReference:
        """reference to template graph processing function"""
        return self._function

    @function.setter
    def function(self, function):
        self._function = self._verify_dict(function, "function", FunctionReference)

    @property
    def source(self) -> DataSource:
        """feature set data source definitions"""
        return self._source

    @source.setter
    def source(self, source: Union[BaseSourceDriver, dict]):
        if isinstance(source, dict):
            kind = source.get("kind", "")
            source = source_kind_to_driver[kind].from_dict(source)
        self._source = source

    @property
    def relations(self) -> dict[str, Entity]:
        """feature set relations dict"""
        return self._relations

    @relations.setter
    def relations(self, relations: dict[str, Entity]):
        for col, ent in relations.items():
            if isinstance(ent, str):
                relations[col] = Entity(ent)
        self._relations = ObjectDict.from_dict({"entity": Entity}, relations, "entity")

    def require_processing(self):
        return len(self._graph.steps) > 0

    def validate_no_processing_for_passthrough(self):
        if self.passthrough and self.require_processing():
            raise mlrun.errors.MLRunInvalidArgumentError(
                "passthrough feature set can not have graph transformations"
            )


class FeatureSetStatus(ModelObj):
    def __init__(
        self,
        state=None,
        targets=None,
        stats=None,
        preview=None,
        function_uri=None,
        run_uri=None,
    ):
        """Feature set status object, containing the current feature-set's status.

        .. warning::
            This class should not be modified directly. It is managed by the parent feature-set object or using
            feature-store APIs. Modifying the status manually may result in unpredictable behaviour.

        :param state: object's current state
        :param targets: list of the data targets used in the last ingestion operation
        :param stats: feature statistics calculated in the last ingestion (if stats calculation was requested)
        :param preview: preview of the feature-set contents (if preview generation was requested)
        :param function_uri: function used to execute the feature-set graph
        :param run_uri: last run used for ingestion
        """

        self.state = state or mlrun.common.schemas.object.ObjectStatusState.CREATED
        self._targets: ObjectList = None
        self.targets = targets or []
        self.stats = stats or {}
        self.preview = preview or []
        self.function_uri = function_uri
        self.run_uri = run_uri

    @property
    def targets(self) -> list[DataTarget]:
        """list of material storage targets + their status/path"""
        return self._targets

    @targets.setter
    def targets(self, targets: list[DataTarget]):
        self._targets = ObjectList.from_list(DataTarget, targets)

    def update_target(self, target: DataTarget):
        self._targets.update(target)

    def update_last_written_for_target(self, target_path: str, last_written: datetime):
        for target in self._targets:
            actual_target_path = get_target_driver(target).get_target_path()
            if (
                actual_target_path == target_path
                or actual_target_path.rstrip("/") == target_path
            ):
                target.last_written = last_written


def emit_policy_to_dict(policy: EmitPolicy):
    # Storey expects the policy to be converted to a dictionary with specific params and won't allow extra params
    # (see Storey's _dict_to_emit_policy function). This takes care of creating a dict conforming to it.
    # TODO - fix Storey's handling of emit policy and parsing of dict in _dict_to_emit_policy.
    struct = {"mode": policy.name()}
    if hasattr(policy, "delay_in_seconds"):
        struct["delay"] = getattr(policy, "delay_in_seconds")
    if hasattr(policy, "max_events"):
        struct["maxEvents"] = getattr(policy, "max_events")
    return struct


class FeatureSet(ModelObj):
    kind = mlrun.common.schemas.ObjectKind.feature_set.value
    _dict_fields = ["kind", "metadata", "spec", "status"]

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        entities: Optional[list[Union[Entity, str]]] = None,
        timestamp_key: Optional[str] = None,
        engine: Optional[str] = None,
        label_column: Optional[str] = None,
        relations: Optional[dict[str, Union[Entity, str]]] = None,
        passthrough: Optional[bool] = None,
    ):
        """Feature set object, defines a set of features and their data pipeline

        example::

            import mlrun.feature_store as fstore

            ticks = fstore.FeatureSet(
                "ticks", entities=["stock"], timestamp_key="timestamp"
            )
            ticks.ingest(df)

        :param name:          name of the feature set
        :param description:   text description
        :param entities:      list of entity (index key) names or :py:class:`~mlrun.features.FeatureSet.Entity`
        :param timestamp_key: timestamp column name
        :param engine:        name of the processing engine (storey, pandas, or spark), defaults to storey
        :param label_column:  name of the label column (the one holding the target (y) values)
        :param relations:     dictionary that indicates all the relations this feature set
                              have with another feature sets. The format of this dictionary is
                              {"my_column":Entity, ...}
        :param passthrough:   if true, ingest will skip offline targets, and get_offline_features will read
                              directly from source
        """
        self._spec: FeatureSetSpec = None
        self._metadata = None
        self._status = None
        self._api_client = None
        self._run_db = None

        self.spec = FeatureSetSpec(
            description=description,
            entities=entities,
            timestamp_key=timestamp_key,
            engine=engine,
            label_column=label_column,
            relations=relations,
            passthrough=passthrough,
        )

        if timestamp_key in self.spec.entities.keys():
            raise mlrun.errors.MLRunInvalidArgumentError(
                "timestamp key can not be entity"
            )

        self.metadata = VersionedObjMetadata(name=name)
        self.status = None
        self._last_state = ""
        self._aggregations = {}
        self.set_targets()

    @property
    def spec(self) -> FeatureSetSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", FeatureSetSpec)

    @property
    def metadata(self) -> VersionedObjMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", VersionedObjMetadata)

    @property
    def status(self) -> FeatureSetStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", FeatureSetStatus)

    @property
    def uri(self):
        """fully qualified feature set uri"""
        return get_store_uri(StorePrefix.FeatureSet, self.fullname)

    @property
    def fullname(self) -> str:
        """full name in the form ``{project}/{name}[:{tag}]``"""
        project = self._metadata.project or mlconf.active_project
        name = self._metadata.name
        tag = self._metadata.tag

        fullname = name
        if project:
            fullname = f"{project}/{fullname}"
        if tag:
            fullname += f":{tag}"
        return fullname

    def _get_run_db(self):
        if self._run_db:
            return self._run_db
        else:
            return mlrun.get_run_db()

    def _override_run_db(self, run_db):
        self._run_db = run_db

    def get_target_path(self, name=None):
        """get the url/path for an offline or specified data target"""
        target = get_offline_target(self, name=name)

        if not target and name:
            target = get_online_target(self, name)

        if target:
            return target.get_path().get_absolute_path(
                project_name=self.metadata.project
            )

    def set_targets(
        self,
        targets=None,
        with_defaults=True,
        default_final_step=None,
    ):
        """set the desired target list or defaults

        :param targets:  list of target type names ('csv', 'nosql', ..) or target objects
                         CSVTarget(), ParquetTarget(), NoSqlTarget(), StreamTarget(), ..
        :param with_defaults: add the default targets (as defined in the central config)
        :param default_final_step: the final graph step after which we add the
                                    target writers, used when the graph branches and
                                    the end cant be determined automatically
        """
        if targets is not None and not isinstance(targets, list):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "targets can only be None or a list of kinds or DataTargetBase derivatives"
            )
        targets = targets or []
        if with_defaults:
            self.spec.with_default_targets = True
            targets.extend(get_default_targets(offline_only=self.spec.passthrough))
        else:
            self.spec.with_default_targets = False

        self.spec.targets = []
        self.__set_targets_add_targets_helper(targets)

        if default_final_step:
            self.spec.graph.final_step = default_final_step

    def __set_targets_add_targets_helper(self, targets):
        """
        Add the desired target list

        :param targets: list of target type names ('csv', 'nosql', ..) or target objects
                         CSVTarget(), ParquetTarget(), NoSqlTarget(), StreamTarget(), ..
        """
        validate_target_list(targets=targets)
        for target in targets:
            kind = target.kind if hasattr(target, "kind") else target
            if kind not in TargetTypes.all():
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"target kind is not supported, use one of: {','.join(TargetTypes.all())}"
                )
            if not hasattr(target, "kind"):
                target = DataTargetBase(
                    target, name=str(target), partitioned=(target == "parquet")
                )
            self.spec.targets.update(target)

    def validate_steps(self, namespace):
        if not self.spec:
            return
        if not self.spec.graph:
            return
        for step in self.spec.graph.steps.values():
            if (
                step.class_name in queue_class_names
                or step.class_name is None
                or "." not in step.class_name
            ):
                #  we are not checking none class names or queue class names.
                continue
            class_object, class_name = step.get_step_class_object(namespace=namespace)
            if not hasattr(class_object, "validate_args"):
                continue
            class_args = step.get_full_class_args(
                namespace=namespace, class_object=class_object
            )
            if class_name.startswith("storey"):
                class_object.validate_args(
                    **(class_args if class_args is not None else {})
                )
            else:
                class_object.validate_args(
                    self, **(class_args if class_args is not None else {})
                )

    def purge_targets(
        self, target_names: Optional[list[str]] = None, silent: bool = False
    ):
        """Delete data of specific targets
        :param target_names: List of names of targets to delete (default: delete all ingested targets)
        :param silent: Fail silently if target doesn't exist in featureset status"""

        verify_feature_set_permissions(
            self, mlrun.common.schemas.AuthorizationAction.delete
        )

        purge_targets = self._reload_and_get_status_targets(
            target_names=target_names, silent=silent
        )

        if purge_targets:
            purge_target_names = list(purge_targets.keys())
            for target_name in purge_target_names:
                target = purge_targets[target_name]
                driver = get_target_driver(target_spec=target, resource=self)
                try:
                    driver.purge()
                except FileNotFoundError:
                    pass
                del self.status.targets[target_name]

            self.save()

    def update_targets_for_ingest(
        self,
        targets: list[DataTargetBase],
        overwrite: Optional[bool] = None,
    ):
        if not targets:
            return

        ingestion_target_names = [t.name for t in targets]

        status_targets = {}
        if not overwrite:
            # silent=True always because targets are not guaranteed to be found in status
            status_targets = (
                self._reload_and_get_status_targets(
                    target_names=ingestion_target_names, silent=True
                )
                or {}
            )

        update_targets_run_id_for_ingest(overwrite, targets, status_targets)

    def _reload_and_get_status_targets(
        self, target_names: Optional[list[str]] = None, silent: bool = False
    ):
        try:
            self.reload(update_spec=False)
        except mlrun.errors.MLRunNotFoundError:
            # If the feature set doesn't exist in DB there shouldn't be any target to delete
            if silent:
                return
            else:
                raise

        if target_names:
            targets = ObjectList(DataTarget)
            for target_name in target_names:
                try:
                    targets[target_name] = self.status.targets[target_name]
                except KeyError:
                    if silent:
                        pass
                    else:
                        raise mlrun.errors.MLRunNotFoundError(
                            f"Target not found in status (fset={self.metadata.name}, target={target_name})"
                        )
        else:
            targets = self.status.targets

        return targets

    def has_valid_source(self):
        """check if object's spec has a valid (non empty) source definition"""
        source = self.spec.source
        return source is not None and source.path is not None and source.path != "None"

    def add_entity(
        self,
        name: str,
        value_type: mlrun.data_types.ValueType = None,
        description: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
    ):
        """add/set an entity (dataset index)

        example::

            import mlrun.feature_store as fstore

            ticks = fstore.FeatureSet(
                "ticks", entities=["stock"], timestamp_key="timestamp"
            )
            ticks.add_entity(
                "country", mlrun.data_types.ValueType.STRING, description="stock country"
            )
            ticks.add_entity("year", mlrun.data_types.ValueType.INT16)
            ticks.save()

        :param name:        entity name
        :param value_type:  type of the entity (default to ValueType.STRING)
        :param description: description of the entity
        :param labels:      label tags dict
        """
        entity = Entity(name, value_type, description=description, labels=labels)
        self._spec.entities.update(entity, name)

    def add_feature(self, feature: mlrun.features.Feature, name=None):
        """add/set a feature

        example::

            import mlrun.feature_store as fstore
            from mlrun.features import Feature

            ticks = fstore.FeatureSet(
                "ticks", entities=["stock"], timestamp_key="timestamp"
            )
            ticks.add_feature(
                Feature(
                    value_type=mlrun.data_types.ValueType.STRING,
                    description="client consistency",
                ),
                "ABC01",
            )
            ticks.add_feature(
                Feature(
                    value_type=mlrun.data_types.ValueType.FLOAT,
                    description="client volatility",
                ),
                "SAB",
            )
            ticks.save()

        :param feature:         setting of Feature
        :param name:            feature name
        """
        self._spec.features.update(feature, name)

    def link_analysis(self, name, uri):
        """add a linked file/artifact (chart, data, ..)"""
        self._spec.analysis[name] = uri

    @property
    def graph(self) -> RootFlowStep:
        """feature set transformation graph/DAG"""
        return self.spec.graph

    def _add_aggregation_to_existing(self, new_aggregation):
        name = new_aggregation["name"]
        if name in self._aggregations:
            current_aggr = self._aggregations[name]
            if current_aggr["windows"] != new_aggregation["windows"]:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Aggregation with name {name} already exists but with window {current_aggr['windows']}. "
                    f"Please provide name for the aggregation"
                )
            if current_aggr["period"] != new_aggregation["period"]:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Aggregation with name {name} already exists but with period {current_aggr['period']}. "
                    f"Please provide name for the aggregation"
                )
            if current_aggr["column"] != new_aggregation["column"]:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Aggregation with name {name} already exists but for different column {current_aggr['column']}. "
                    f"Please provide name for the aggregation"
                )
            current_aggr["operations"] = list(
                set(current_aggr["operations"] + new_aggregation["operations"])
            )

            return
        self._aggregations[name] = new_aggregation

    def add_aggregation(
        self,
        column,
        operations,
        windows,
        period=None,
        name=None,
        step_name=None,
        after=None,
        before=None,
        emit_policy: EmitPolicy = None,
    ):
        """add feature aggregation rule

        example::

            myset.add_aggregation("ask", ["sum", "max"], "1h", "10m", name="asks")

        :param column:     name of column/field aggregate. Do not name columns starting with either `_` or `aggr_`.
                           They are reserved for internal use, and the data does not ingest correctly.
                           When using the pandas engine, do not use spaces (` `) or periods (`.`) in the column names;
                           they cause errors in the ingestion.
        :param operations: aggregation operations. Supported operations:
                             count, sum, sqr, max, min, first, last, avg, stdvar, stddev
        :param windows:    time windows, can be a single window, e.g. '1h', '1d',
                            or a list of same unit windows e.g. ['1h', '6h']
                            windows are transformed to fixed windows or
                            sliding windows depending whether period parameter
                            provided.

                            - Sliding window is fixed-size overlapping windows
                              that slides with time.
                              The window size determines the size of the sliding window
                              and the period determines the step size to slide.
                              Period must be integral divisor of the window size.
                              If the period is not provided then fixed windows is used.

                            - Fixed window is fixed-size, non-overlapping, gap-less window.
                              The window is referred to as a tumbling window.
                              In this case, each record on an in-application stream belongs
                              to a specific window. It is processed only once
                              (when the query processes the window to which the record belongs).
        :param period:     optional, sliding window granularity, e.g. '20s' '10m'  '3h' '7d'
        :param name:       optional, aggregation name/prefix. Must be unique per feature set. If not passed,
                            the column will be used as name.
        :param step_name: optional, graph step name
        :param after:      optional, after which graph step it runs
        :param before:     optional, comes before graph step
        :param emit_policy: optional, which emit policy to use when performing the aggregations. Use the derived
                            classes of ``storey.EmitPolicy``. The default is to emit every period for Spark engine
                            and emit every event for storey. Currently the only other supported option is to use
                            ``emit_policy=storey.EmitEveryEvent()`` when using the Spark engine to emit every event

        """
        if isinstance(operations, str):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Invalid parameters provided - operations must be a list."
            )

        name = name or column

        if isinstance(windows, str):
            windows = [windows]
        if isinstance(operations, str):
            operations = [operations]

        aggregation = FeatureAggregation(
            name, column, operations, windows, period
        ).to_dict()

        def upsert_feature(name):
            if name in self.spec.features:
                self.spec.features[name].aggregate = True
            else:
                self.spec.features[name] = Feature(
                    name=column, aggregate=True, value_type="float"
                )

        step_name = step_name or aggregates_step
        graph = self.spec.graph
        if step_name in graph.steps:
            step = graph.steps[step_name]
            self._add_aggregation_to_existing(aggregation)
            step.class_args["aggregates"] = list(self._aggregations.values())
            if emit_policy and self.spec.engine == "spark":
                # Using simple override here - we might want to consider exploding if different emit policies
                # were used for multiple aggregations.
                emit_policy_dict = emit_policy_to_dict(emit_policy)
                if "emit_policy" in step.class_args:
                    curr_emit_policy = step.class_args["emit_policy"]["mode"]
                    if curr_emit_policy != emit_policy_dict["mode"]:
                        logger.warning(
                            f"Current emit policy will be overridden: {curr_emit_policy} => {emit_policy_dict['mode']}"
                        )
                step.class_args["emit_policy"] = emit_policy_dict
        else:
            class_args = {}
            self._aggregations[aggregation["name"]] = aggregation
            if before is None and after is None:
                after = previous_step
            if not self.spec.engine or self.spec.engine == "storey":
                step = graph.add_step(
                    name=step_name,
                    after=after,
                    before=before,
                    class_name="storey.AggregateByKey",
                    time_field=self.spec.timestamp_key,
                    aggregates=[aggregation],
                    table=".",
                    **class_args,
                )
            elif self.spec.engine == "spark":
                key_columns = []
                if emit_policy:
                    class_args["emit_policy"] = emit_policy_to_dict(emit_policy)
                for entity in self.spec.entities:
                    key_columns.append(entity.name)
                step = graph.add_step(
                    name=step_name,
                    key_columns=key_columns,
                    time_column=self.spec.timestamp_key,
                    aggregates=[aggregation],
                    after=after,
                    before=before,
                    class_name="mlrun.feature_store.feature_set.SparkAggregateByKey",
                    **class_args,
                )
            else:
                raise ValueError(
                    "Aggregations are only implemented for storey and spark engines."
                )

        for operation in operations:
            for window in windows:
                upsert_feature(f"{name}_{operation}_{window}")

        return step

    def get_stats_table(self):
        """get feature statistics table (as dataframe)"""
        if self.status.stats:
            return pd.DataFrame.from_dict(self.status.stats, orient="index")

    def __getitem__(self, name):
        return self._spec.features[name]

    def __setitem__(self, key, item):
        if key not in self._spec.entities.keys():
            self._spec.features.update(item, key)
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "A `FeatureSet` cannot have an entity and a feature with the same name. "
                f"The feature that was given to add '{key}' has the same name of the `FeatureSet`'s entity."
            )

    def plot(self, filename=None, format=None, with_targets=False, **kw):
        """plot/save graph using graphviz

        example::

            import mlrun.feature_store as fstore

            ...
            ticks = fstore.FeatureSet(
                "ticks", entities=["stock"], timestamp_key="timestamp"
            )
            ticks.add_aggregation(
                name="priceN",
                column="price",
                operations=["avg"],
                windows=["1d"],
                period="1h",
            )
            ticks.plot(rankdir="LR", with_targets=True)

        :param filename:     target filepath for the graph image (None for the notebook)
        :param format:       the output format used for rendering (``'pdf'``, ``'png'``, etc.)
        :param with_targets: show targets in the graph image
        :param kw:           kwargs passed to graphviz, e.g. rankdir=”LR” (see https://graphviz.org/doc/info/attrs.html)
        :return:             graphviz graph object
        """
        graph = self.spec.graph
        _, default_final_step, _ = graph.check_and_process_graph(allow_empty=True)
        targets = None
        if with_targets:
            validate_target_list(targets=targets)
            validate_target_placement(graph, default_final_step, self.spec.targets)
            targets = [
                BaseStep(
                    f"{target.kind}/{target.name}",
                    after=target.after_step or default_final_step,
                    shape="cylinder",
                )
                for target in self.spec.targets
            ]
        return graph.plot(filename, format, targets=targets, **kw)

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        target_name=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
        **kwargs,
    ):
        """return featureset (offline) data as dataframe

        :param columns:      list of columns to select (if not all)
        :param df_module:    py module used to create the DataFrame (pd for Pandas, dd for Dask, ..)
        :param target_name:  select a specific target (material view)
        :param start_time:   filter by start time
        :param end_time:     filter by end time
        :param time_column:  specify the time column name in the file
        :param kwargs:       additional reader (csv, parquet, ..) args
        :param additional_filters: List of additional_filter conditions as tuples.
                                    Each tuple should be in the format (column_name, operator, value).
                                    Supported operators: "=", ">=", "<=", ">", "<".
                                    Example: [("Product", "=", "Computer")]
                                    For all supported filters, please see:
                                    https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
        :return: DataFrame
        """
        entities = list(self.spec.entities.keys())
        if columns:
            if self.spec.timestamp_key and self.spec.timestamp_key not in entities:
                columns = [self.spec.timestamp_key] + columns
            columns = entities + columns

        if self.spec.passthrough:
            if not self.spec.source:
                raise mlrun.errors.MLRunNotFoundError(
                    "passthrough feature set {self.metadata.name} with no source"
                )
            df = self.spec.source.to_dataframe(
                columns=columns,
                start_time=start_time,
                end_time=end_time,
                time_field=time_column,
                additional_filters=additional_filters,
                **kwargs,
            )
            # to_dataframe() can sometimes return an iterator of dataframes instead of one dataframe
            if not isinstance(df, pd.DataFrame):
                df = pd.concat(df)
            return df

        target = get_offline_target(self, name=target_name)
        if not target:
            raise mlrun.errors.MLRunNotFoundError(
                "there are no offline targets for this feature set"
            )
        result = target.as_df(
            columns=columns,
            df_module=df_module,
            entities=entities,
            start_time=start_time,
            end_time=end_time,
            time_column=time_column,
            additional_filters=additional_filters,
            **kwargs,
        )
        return result

    def save(self, tag="", versioned=False):
        """save to mlrun db"""
        db = self._get_run_db()
        self.metadata.project = self.metadata.project or mlconf.active_project
        tag = tag or self.metadata.tag or "latest"
        as_dict = self.to_dict()
        as_dict["spec"]["features"] = as_dict["spec"].get(
            "features", []
        )  # bypass DB bug
        db.store_feature_set(as_dict, tag=tag, versioned=versioned)

    def reload(self, update_spec=True):
        """reload/sync the feature vector status and spec from the DB"""
        feature_set = self._get_run_db().get_feature_set(
            self.metadata.name, self.metadata.project, self.metadata.tag
        )
        if isinstance(feature_set, dict):
            feature_set = FeatureSet.from_dict(feature_set)

        self.status = feature_set.status
        if update_spec:
            self.spec = feature_set.spec

    def ingest(
        self,
        source=None,
        targets: Optional[list[DataTargetBase]] = None,
        namespace=None,
        return_df: bool = True,
        infer_options: InferOptions = InferOptions.default(),
        run_config: RunConfig = None,
        mlrun_context=None,
        spark_context=None,
        overwrite=None,
    ) -> Optional[pd.DataFrame]:
        """Read local DataFrame, file, URL, or source into the feature store
        Ingest reads from the source, run the graph transformations, infers  metadata and stats
        and writes the results to the default of specified targets

        when targets are not specified data is stored in the configured default targets
        (will usually be NoSQL for real-time and Parquet for offline).

        the `run_config` parameter allow specifying the function and job configuration,
        see: :py:class:`~mlrun.feature_store.RunConfig`

        example::

            stocks_set = FeatureSet("stocks", entities=[Entity("ticker")])
            stocks = pd.read_csv("stocks.csv")
            df = stocks_set.ingest(stocks, infer_options=fstore.InferOptions.default())

            # for running as remote job
            config = RunConfig(image="mlrun/mlrun")
            df = ingest(stocks_set, stocks, run_config=config)

            # specify source and targets
            source = CSVSource("mycsv", path="measurements.csv")
            targets = [CSVTarget("mycsv", path="./mycsv.csv")]
            ingest(measurements, source, targets)

        :param source:        source dataframe or other sources (e.g. parquet source see:
                            :py:class:`~mlrun.datastore.ParquetSource` and other classes in mlrun.datastore with suffix
                            Source)
        :param targets:       optional list of data target objects
        :param namespace:     namespace or module containing graph classes
        :param return_df:     indicate if to return a dataframe with the graph results
        :param infer_options: schema (for discovery of entities, features in featureset), index, stats,
                            histogram and preview infer options (:py:class:`~mlrun.feature_store.InferOptions`)
        :param run_config:    function and/or run configuration for remote jobs,
                            see :py:class:`~mlrun.feature_store.RunConfig`
        :param mlrun_context: mlrun context (when running as a job), for internal use !
        :param spark_context: local spark session for spark ingestion, example for creating the spark context:
                            `spark = SparkSession.builder.appName("Spark function").getOrCreate()`
                            For remote spark ingestion, this should contain the remote spark service name
        :param overwrite:     delete the targets' data prior to ingestion
                            (default: True for non scheduled ingest - deletes the targets that are about to be ingested.
                            False for scheduled ingest - does not delete the target)
        :return:              if return_df is True, a dataframe will be returned based on the graph
        """
        return mlrun.feature_store.api._ingest(
            self,
            source,
            targets,
            namespace,
            return_df,
            infer_options,
            run_config,
            mlrun_context,
            spark_context,
            overwrite,
        )

    def preview(
        self,
        source,
        entity_columns: Optional[list] = None,
        namespace=None,
        options: InferOptions = None,
        verbose: bool = False,
        sample_size: Optional[int] = None,
    ) -> pd.DataFrame:
        """run the ingestion pipeline with local DataFrame/file data and infer features schema and stats

        example::

            quotes_set = FeatureSet("stock-quotes", entities=[Entity("ticker")])
            quotes_set.add_aggregation("ask", ["sum", "max"], ["1h", "5h"], "10m")
            quotes_set.add_aggregation("bid", ["min", "max"], ["1h"], "10m")
            df = quotes_set.preview(
                quotes_df,
                entity_columns=["ticker"],
            )

        :param source:         source dataframe or csv/parquet file path
        :param entity_columns: list of entity (index) column names
        :param namespace:      namespace or module containing graph classes
        :param options:        schema (for discovery of entities, features in featureset), index, stats,
                            histogram and preview infer options (:py:class:`~mlrun.feature_store.InferOptions`)
        :param verbose:        verbose log
        :param sample_size:    num of rows to sample from the dataset (for large datasets)
        """
        return mlrun.feature_store.api._preview(
            self, source, entity_columns, namespace, options, verbose, sample_size
        )

    def deploy_ingestion_service(
        self,
        source: DataSource = None,
        targets: Optional[list[DataTargetBase]] = None,
        name: Optional[str] = None,
        run_config: RunConfig = None,
        verbose=False,
    ) -> tuple[str, BaseRuntime]:
        """Start real-time ingestion service using nuclio function

        Deploy a real-time function implementing feature ingestion pipeline
        the source maps to Nuclio event triggers (http, kafka, v3io stream, etc.)

        the `run_config` parameter allow specifying the function and job configuration,
        see: :py:class:`~mlrun.feature_store.RunConfig`

        example::

            source = HTTPSource()
            func = mlrun.code_to_function("ingest", kind="serving").apply(mount_v3io())
            config = RunConfig(function=func)
            my_set.deploy_ingestion_service(source, run_config=config)

        :param source:        data source object describing the online or offline source
        :param targets:       list of data target objects
        :param name:          name for the job/function
        :param run_config:    service runtime configuration (function object/uri, resources, etc..)
        :param verbose:       verbose log

        :return: URL to access the deployed ingestion service, and the function that was deployed (which will
                differ from the function passed in via the run_config parameter).
        """

        return mlrun.feature_store.api._deploy_ingestion_service_v2(
            self, source, targets, name, run_config, verbose
        )

    def extract_relation_keys(
        self,
        other_feature_set,
        relations: Optional[dict[str, Union[str, Entity]]] = None,
    ) -> list[str]:
        """
        Checks whether a feature set can be merged to the right of this feature set.

        :param other_feature_set:   The feature set to be merged to the right of this feature set.
        :param relations:           The relations that were defined on this feature set.
        :returns:                   If the two feature sets can be merged, a list of the left join keys is returned.
                                    Otherwise, an empty list is returned.
                                    (The right join keys are always the entities of the other feature set)
        """
        right_feature_set_entity_list = other_feature_set.spec.entities

        if all(
            ent in self.spec.entities for ent in right_feature_set_entity_list
        ) and len(right_feature_set_entity_list) == len(self.spec.entities):
            # entities wise
            return list(self.spec.entities.keys())
        elif all(ent in self.spec.entities for ent in right_feature_set_entity_list):
            # entities wise when the right fset have lower number of entities
            return list(right_feature_set_entity_list.keys())

        elif relations:
            curr_col_relations_list = list(
                map(
                    lambda ent: (
                        list(relations.keys())[list(relations.values()).index(ent)]
                        if ent in relations.values()
                        else False
                    ),
                    right_feature_set_entity_list,
                )
            )

            if all(curr_col_relations_list):
                return curr_col_relations_list

        return []

    def is_connectable_to_df(self, df_columns: list[str]) -> bool:
        """
        This method checks if the dataframe can be left-joined with this feature set

        :param df_columns:  The columns of the data frame you want to merge to the left of this feature set
        :return:            True if it can be left-joined and False otherwise
        """
        return df_columns and all(
            ent in df_columns for ent in self.spec.entities.keys()
        )


class SparkAggregateByKey(StepToDict):
    _supported_operations = [
        "count",
        "sum",
        "sqr",
        "max",
        "min",
        "first",
        "last",
        "avg",
        "stdvar",
        "stddev",
    ]

    def __init__(
        self,
        key_columns: list[str],
        time_column: str,
        aggregates: list[dict],
        emit_policy: Union[EmitPolicy, dict] = None,
    ):
        self.key_columns = key_columns
        self.time_column = time_column
        self.aggregates = aggregates
        self.emit_policy_mode = None
        if emit_policy:
            if isinstance(emit_policy, EmitPolicy):
                emit_policy = emit_policy_to_dict(emit_policy)
            self.emit_policy_mode = emit_policy["mode"]

    @staticmethod
    def _duration_to_spark_format(duration):
        num = duration[:-1]
        unit = duration[-1:]
        if unit == "d":
            unit = "day"
        elif unit == "h":
            unit = "hour"
        elif unit == "m":
            unit = "minute"
        elif unit == "s":
            unit = "second"
        else:
            raise ValueError(f"Invalid duration '{duration}'")
        return f"{num} {unit}"

    @staticmethod
    def _verify_operation(op):
        if op not in SparkAggregateByKey._supported_operations:
            error_string = (
                f"operation {op} is unsupported. Supported operations: "
                + ", ".join(SparkAggregateByKey._supported_operations)
            )
            raise mlrun.errors.MLRunInvalidArgumentError(error_string)

    def _extract_fields_from_aggregate_dict(self, aggregate):
        name = aggregate["name"]
        column = aggregate["column"]
        operations = aggregate["operations"]
        for op in operations:
            self._verify_operation(op)
        windows = aggregate["windows"]
        spark_period = (
            self._duration_to_spark_format(aggregate["period"])
            if "period" in aggregate
            else None
        )
        return name, column, operations, windows, spark_period

    @staticmethod
    def _get_aggr(operation, column):
        import pyspark.sql.functions as funcs

        if operation == "sqr":
            return funcs.sum(funcs.expr(f"{column} * {column}"))
        elif operation == "stdvar":
            return funcs.variance(column)
        else:
            func = getattr(funcs, operation)
            return func(column)

    def do(self, event):
        import pyspark.sql.functions as funcs
        from pyspark.sql import Window

        time_column = self.time_column or "time"
        input_df = event

        if not self.emit_policy_mode or self.emit_policy_mode != EmitEveryEvent.name():
            last_value_aggs = [
                funcs.last(column).alias(column)
                for column in input_df.columns
                if column not in self.key_columns and column != time_column
            ]

            dfs = []
            for aggregate in self.aggregates:
                (
                    name,
                    column,
                    operations,
                    windows,
                    spark_period,
                ) = self._extract_fields_from_aggregate_dict(aggregate)

                for window in windows:
                    spark_window = self._duration_to_spark_format(window)
                    aggs = last_value_aggs
                    for operation in operations:
                        agg = self._get_aggr(operation, column)
                        agg_name = f"{name if name else column}_{operation}_{window}"
                        agg = agg.alias(agg_name)
                        aggs.append(agg)
                    window_column = funcs.window(
                        time_column, spark_window, spark_period
                    )
                    df = input_df.groupBy(
                        *self.key_columns,
                        window_column.end.alias(time_column),
                    ).agg(*aggs)
                    df = df.withColumn(f"{time_column}_window", funcs.lit(window))
                    dfs.append(df)

            union_df = dfs[0]
            for df in dfs[1:]:
                union_df = union_df.unionByName(df, allowMissingColumns=True)

            return union_df

        else:
            window_counter = 0
            # We'll use this column to identify our original row and group-by across the various windows
            # (either sliding windows or multiple windows provided). See below comment for more details.
            rowid_col = "__mlrun_rowid"
            df = input_df.withColumn(rowid_col, funcs.monotonically_increasing_id())

            drop_columns = [rowid_col]
            window_rank_cols = []
            union_df = None
            for aggregate in self.aggregates:
                (
                    name,
                    column,
                    operations,
                    windows,
                    spark_period,
                ) = self._extract_fields_from_aggregate_dict(aggregate)

                for window in windows:
                    spark_window = self._duration_to_spark_format(window)
                    window_col = f"__mlrun_window_{window_counter}"
                    win_df = df.withColumn(
                        window_col,
                        funcs.window(time_column, spark_window, spark_period).end,
                    )
                    function_window = Window.partitionBy(*self.key_columns, window_col)

                    window_rank_col = f"__mlrun_win_rank_{window_counter}"
                    rank_window = Window.partitionBy(rowid_col).orderBy(window_col)
                    win_df = win_df.withColumn(
                        window_rank_col, funcs.row_number().over(rank_window)
                    )
                    window_rank_cols.append(window_rank_col)
                    drop_columns.extend([window_col, window_rank_col])

                    window_counter += 1

                    for operation in operations:
                        agg = self._get_aggr(operation, column)
                        agg_name = f"{name if name else column}_{operation}_{window}"
                        win_df = win_df.withColumn(agg_name, agg.over(function_window))

                    union_df = (
                        union_df.unionByName(win_df, allowMissingColumns=True)
                        if union_df
                        else win_df
                    )

            # We need to collapse the multiple window rows that were generated during the query processing. For that
            # purpose we'll pick just the 1st row for each window, and then group-by with ignorenulls. Basically since
            # the result is a union of multiple windows, we'll get something like this for each input row:
            # row   window_1    rank_window_1   window_2    rank_window_2   ...calculations and fields...
            # 1     10:00       1               null        null            ...
            # 2     10:10       2               null        null            ...
            # 3     null        null            10:00       1               ...
            # 4     null        null            10:10       2               ...
            # And we want to take rows 1 and 3 in this case. Then the group-by will merge them to a single line since
            # it ignores nulls, so it will take the values for window_1 from row 1 and for window_2 from row 3.
            window_filter = " or ".join(
                [f"{window_rank_col} == 1" for window_rank_col in window_rank_cols]
            )
            first_value_aggs = [
                funcs.first(column, ignorenulls=True).alias(column)
                for column in union_df.columns
                if column not in drop_columns
            ]

            return (
                union_df.filter(window_filter)
                .groupBy(rowid_col)
                .agg(*first_value_aggs)
                .drop(rowid_col)
            )



# Copyright 2018 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Optional, Union

from mlrun.artifacts import Artifact, ModelArtifact
from mlrun.execution import MLClientCtx
from mlrun.model import RunObject
from mlrun.projects import MlrunProject


class Tracker(ABC):
    """
    Abstract tracker class, describes the interface for a tracker: A class for tracking 3rd party vendor's for logging
    their artifacts and experiments into MLRun. There are two tracking modes:

    * Online: While running an MLRun function, a tracker will be used via the `pre_run` and `post_run` methods to track
      a run from the tracked vendor into MLRun.
    * Offline: Manually importing models and artifacts into an MLRun project using the `import_x` methods.
    """

    @staticmethod
    @abstractmethod
    def is_enabled() -> bool:
        """
        Checks if tracker is enabled.

        :return: True if the tracking configuration is enabled, False otherwise.
        """
        pass

    def pre_run(self, context: MLClientCtx):
        """
        Initializes the tracking system for a 3rd party module. This function sets up the necessary components and
        resources to enable tracking of the module.

        :param context: Current mlrun context
        """
        pass

    def post_run(self, context: Union[MLClientCtx, dict]):
        """
        Performs post-run tasks for logging 3rd party artifacts generated during the run.

        :param context: Current mlrun context
        """
        pass

    def import_run(
        self,
        project: MlrunProject,
        reference_id: Any,
        function_name: str,
        handler: Optional[str] = None,
        **kwargs,
    ) -> RunObject:
        """
        Import a previous run from a 3rd party vendor to MLRun.

        :param project:       The MLRun project to import the run to.
        :param reference_id:  A reference for the run to import from the 3rd party vendor.
        :param function_name: The MLRun function to assign this run to.
        :param handler:       The handler for MLRun's RunObject

        :return: The newly imported run object.
        """
        pass

    def import_model(
        self, project: MlrunProject, reference_id: Any, **kwargs
    ) -> ModelArtifact:
        """
        Import a model from a 3rd party vendor to MLRun.

        :param project:      The MLRun project to import the model to.
        :param reference_id: A reference for the model to import from the 3rd party vendor.

        :return: The newly imported model artifact.
        """
        pass

    def import_artifact(
        self, project: MlrunProject, reference_id: Any, **kwargs
    ) -> Artifact:
        """
        Import an artifact from a 3rd party vendor to MLRun.

        :param project:      The MLRun project to import the artifact to.
        :param reference_id: A reference for the artifact to import from the 3rd party vendor.

        :return: The newly imported artifact.
        """
        pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.common.types


class ListRuntimeResourcesGroupByField(mlrun.common.types.StrEnum):
    job = "job"
    project = "project"


class RuntimeResource(pydantic.v1.BaseModel):
    name: str
    labels: dict[str, str] = {}
    status: typing.Optional[dict]


class RuntimeResources(pydantic.v1.BaseModel):
    crd_resources: list[RuntimeResource] = []
    pod_resources: list[RuntimeResource] = []
    # only for dask runtime
    service_resources: typing.Optional[list[RuntimeResource]] = None

    class Config:
        extra = pydantic.v1.Extra.allow


class KindRuntimeResources(pydantic.v1.BaseModel):
    kind: str
    resources: RuntimeResources


RuntimeResourcesOutput = list[KindRuntimeResources]


# project name -> job uid -> runtime resources
GroupedByJobRuntimeResourcesOutput = dict[str, dict[str, RuntimeResources]]
# project name -> kind -> runtime resources
GroupedByProjectRuntimeResourcesOutput = dict[str, dict[str, RuntimeResources]]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Callable, Optional, Union

import tensorflow as tf
from packaging import version
from tensorboard.plugins.hparams import api as hp_api
from tensorboard.plugins.hparams import api_pb2 as hp_api_pb2
from tensorboard.plugins.hparams import summary as hp_summary
from tensorflow import Tensor, Variable
from tensorflow.keras import Model
from tensorflow.python.ops import summary_ops_v2

import mlrun

from ..._dl_common.loggers import TensorboardLogger
from ..utils import TFKerasTypes
from .logging_callback import LoggingCallback


class _TFKerasTensorboardLogger(TensorboardLogger):
    """
    The keras framework implementation of the 'TensorboardLogger'.
    """

    def __init__(
        self,
        statistics_functions: list[Callable[[Union[Variable]], Union[float, Variable]]],
        context: mlrun.MLClientCtx = None,
        tensorboard_directory: Optional[str] = None,
        run_name: Optional[str] = None,
        update_frequency: Union[int, str] = "epoch",
    ):
        """
        Initialize a tensorboard logger callback with the given configuration. At least one of 'context' and
        'tensorboard_directory' must be given.

        :param statistics_functions:  A list of statistics functions to calculate at the end of each epoch on the
                                      tracked weights. Only relevant if weights are being tracked. The functions in
                                      the list must accept one Weight and return a float (or float convertible) value.
        :param context:               A MLRun context to use for logging into the user's tensorboard directory. The
                                      context parameters can be logged as static hyperparameters as well.
        :param tensorboard_directory: If context is not given, or if wished to set the directory even with context,
                                      this will be the output for the event logs of tensorboard. If not given, context
                                      must be provided as the default tensorboard output directory will be:
                                      /User/.tensorboard/<PROJECT_NAME> or if working on local, the set artifacts path.
        :param run_name:              This experiment run name. Each run name will be indexed at the end of the name so
                                      each experiment will be numbered automatically. If a context was given, the
                                      context's uid will be added instead of an index. If a run name was not given the
                                      current time stamp will be used.
        :param update_frequency:      Per how many iterations (batches) the callback should write the tracked values to
                                      tensorboard. Can be passed as a string equal to 'epoch' for per epoch and 'batch'
                                      for per single batch, or as an integer specifying per how many iterations to
                                      update. Notice that writing to tensorboard too frequently may cause the training
                                      to be slower. Default: 'epoch'.
        """
        super().__init__(
            statistics_functions=statistics_functions,
            context=context,
            tensorboard_directory=tensorboard_directory,
            run_name=run_name,
            update_frequency=update_frequency,
        )

        # Setup the tensorboard writer property:
        self._file_writer = None

    def write_model_to_tensorboard(self, model: Model):
        """
        Write the given model as a graph in tensorboard.

        :param model: The model to write to tensorboard.
        """
        with self._file_writer.as_default():
            # Log the model's graph according to tensorflow's version:
            if version.parse(tf.__version__) < version.parse("2.5.0"):
                with summary_ops_v2.always_record_summaries():
                    summary_ops_v2.keras_model(name=model.name, data=model, step=0)
            else:
                from tensorflow.python.keras.callbacks import keras_model_summary

                with summary_ops_v2.record_if(True):
                    keras_model_summary("keras", model, step=0)

    def write_parameters_table_to_tensorboard(self):
        """
        Write the summaries, static and dynamic hyperparameters to the table in tensorboard's hparams section. This
        method is called once for creating the hparams table.
        """
        # Check if needed to track hyperparameters:
        if (
            len(self._static_hyperparameters) == 0
            and len(self._dynamic_hyperparameters) == 0
        ):
            return

        # Prepare the static hyperparameters values:
        non_graph_parameters = {"Date": str(datetime.now()).split(".")[0]}
        hp_param_list = [hp_api.HParam("Date")]
        for parameter, value in self._static_hyperparameters.items():
            non_graph_parameters[parameter] = value
            hp_param_list.append(hp_api.HParam(parameter))

        # Prepare the summaries values and the dynamic hyperparameters values (both registered as metrics):
        graph_parameters = {}
        hp_metric_list = []
        for metric in self._training_results:
            for prefix in ["training", "validation"]:
                metric_name = f"{self._Sections.SUMMARY}/{prefix}_{metric}"
                graph_parameters[metric_name] = 0.0
                hp_metric_list.append(hp_api.Metric(metric_name))
        for parameter, epochs in self._dynamic_hyperparameters.items():
            parameter_name = f"{self._Sections.HYPERPARAMETERS}/{parameter}"
            graph_parameters[parameter_name] = epochs[-1]
            hp_metric_list.append(hp_api.Metric(parameter_name))

        # Write the hyperparameters and summaries to the table:
        with self._file_writer.as_default():
            hp_api.hparams_config(hparams=hp_param_list, metrics=hp_metric_list)
            hp_api.hparams(non_graph_parameters, trial_id=self._run_name)

    def open(self):
        """
        Create the output path and initialize the tensorboard file writer.
        """
        # Create the output path:
        self._create_output_path()

        # Use the output path to initialize the tensorboard file writer:
        self._file_writer = tf.summary.create_file_writer(self._output_path)
        self._file_writer.set_as_default()

    def flush(self):
        """
        Make sure all values were written to the directory logs so it will be available live.
        """
        self._file_writer.flush()

    def close(self):
        """
        Close the file writer object, wrapping up the hyperparameters table.
        """
        # Close the hyperparameters writing:
        if not (
            len(self._static_hyperparameters) == 0
            and len(self._dynamic_hyperparameters) == 0
        ):
            with self._file_writer.as_default():
                pb = hp_summary.session_end_pb(hp_api_pb2.STATUS_SUCCESS)
                raw_pb = pb.SerializeToString()
                tf.compat.v2.summary.experimental.write_raw_pb(raw_pb, step=0)

        # Flush and close the writer:
        self.flush()
        self._file_writer.close()

    def _write_text_to_tensorboard(self, tag: str, text: str, step: int):
        """
        Write text to tensorboard's text section. Summary information of this training / validation run will be logged
        to tensorboard using this method.

        :param tag:  The tag of the text (box it will be appearing under).
        :param text: The text to write.
        :param step: The iteration / epoch the text belongs to.
        """
        with self._file_writer.as_default():
            tf.summary.text(
                name=tag,
                data=text,
                step=step,
            )

    def _write_scalar_to_tensorboard(self, name: str, value: float, step: int):
        """
        Write the scalar's value into its plot.

        :param name:  The plot's name.
        :param value: The value to add to the plot.
        :param step:  The iteration / epoch the value belongs to.
        """
        with self._file_writer.as_default():
            tf.summary.scalar(
                name=name,
                data=value,
                step=step,
            )

    def _write_weight_histogram_to_tensorboard(
        self, name: str, weight: Variable, step: int
    ):
        """
        Write the current state of the weights as histograms to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its histogram.
        :param step:   The iteration / epoch the weight's histogram state belongs to.
        """
        with self._file_writer.as_default():
            tf.summary.histogram(
                name=name,
                data=weight,
                step=step,
            )

    def _write_weight_image_to_tensorboard(
        self, name: str, weight: Variable, step: int
    ):
        """
        Log the current state of the weights as images to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its image.
        :param step:   The iteration / epoch the weight's image state belongs to.
        """
        raise NotImplementedError


class TensorboardLoggingCallback(LoggingCallback):
    """
    Callback for logging data during training / evaluation to tensorboard. the available data in tensorboard will be:

    * Summary text of the run with a hyperlink to the MLRun log if it was done.
    * Hyperparameters tuning table: static hyperparameters, dynamic hyperparameters and epoch validation summaries.
    * Plots:

      * Per iteration (batch) plot for the training and validation metrics.
      * Per epoch plot for the dynamic hyperparameters and validation summaries results.
      * Per epoch weights statistics for each weight and statistic.

    * Histograms per epoch for each of the logged weights.
    * Distributions per epoch for each of the logged weights.
    * Images per epoch for each of the logged weights.
    * Model architecture graph.

    All the collected data will be available in this callback post the training / validation process and can be accessed
    via the 'training_results', 'validation_results', 'static_hyperparameters', 'dynamic_hyperparameters', 'summaries',
    'weights', 'weights_mean' and 'weights_std' properties.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        tensorboard_directory: Optional[str] = None,
        run_name: Optional[str] = None,
        weights: Union[bool, list[str]] = False,
        statistics_functions: Optional[
            list[Callable[[Union[Variable, Tensor]], Union[float, Tensor]]]
        ] = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                Union[list[Union[str, int]], Callable[[], TFKerasTypes.TrackableType]],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[str, Union[TFKerasTypes.TrackableType, list[Union[str, int]]]]
        ] = None,
        update_frequency: Union[int, str] = "epoch",
        auto_log: bool = False,
    ):
        """
        Initialize a tensorboard logging callback with the given weights, hyperparameters and logging configurations.
        Note that at least one of 'context' and 'tensorboard_directory' must be given.

        :param context:                 A mlrun context to use for logging into the user's tensorboard directory.
        :param tensorboard_directory:   If context is not given, or if wished to set the directory even with context,
                                        this will be the output for the event logs of tensorboard. If not given, the
                                        'tensorboard_dir' parameter will be tried to be taken from the provided context.
                                        If not found in the context, the default tensorboard output directory will be:
                                        /User/.tensorboard/<PROJECT_NAME> or if working on local, the set artifacts
                                        path.
        :param run_name:                This experiment run name. Each run name will be indexed at the end of the name
                                        so each experiment will be numbered automatically. If a context was given, the
                                        context's uid will be added instead of an index. If a run name was not given the
                                        current time in the following format: 'YYYY-mm-dd_HH:MM:SS'.
        :param weights:                 If wished to track weights to draw their histograms and calculate statistics per
                                        epoch, the weights names should be passed here. Note that each name given will
                                        be searched as 'if <NAME> in <WEIGHT_NAME>' so a simple module name will be
                                        enough to catch his weights. A boolean value can be passed to track all weights.
                                        Default: False.
        :param statistics_functions:    A list of statistics functions to calculate at the end of each epoch on the
                                        tracked weights. Only relevant if weights are being tracked. The functions in
                                        the list must accept one Parameter (or Tensor) and return a float (or float
                                        convertible) value. The default statistics are 'mean' and 'std'. To get the
                                        default functions list for appending additional functions you can access it via
                                        'TensorboardLoggingCallback.get_default_weight_statistics_list()'. To not track
                                        statistics at all simply pass an empty list '[]'.
        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each epoch) it should
                                        be passed here. The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are a key chain from the model. A key
                                        chain is a list of keys and indices to know how to access the needed
                                        hyperparameter from the model. If the hyperparameter is not of accessible from
                                        the model, a custom callable method can be passed. For example, to track the
                                        'lr' attribute of an optimizer and a custom parameter, one should pass:
                                        {
                                            "learning rate": ["optimizer", "lr"],
                                            "custom_parameter": get_custom_parameter
                                        }
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are the hyperparameter static value
                                        or a key chain just like the dynamic hyperparameter. For example, to track the
                                        'epochs' of an experiment run, one should pass:
                                        {
                                            "epochs": 7
                                        }
        :param update_frequency:        Per how many iterations (batches) the callback should write the tracked values
                                        to tensorboard. Can be passed as a string equal to 'epoch' for per epoch and
                                        'batch' for per single batch, or as an integer specifying per how many
                                        iterations to update. Notice that writing to tensorboard too frequently may
                                        cause the training to be slower. Default: 'epoch'.
        :param auto_log:                Whether or not to enable auto logging for logging the context parameters and
                                        trying to track common static and dynamic hyperparameters such as learning rate.

        :raise MLRunInvalidArgumentError: In case both 'context' and 'tensorboard_directory' parameters were not given
                                          or the 'update_frequency' was incorrect.
        """
        super().__init__(
            dynamic_hyperparameters=dynamic_hyperparameters,
            static_hyperparameters=static_hyperparameters,
            auto_log=auto_log,
        )

        # Replace the logger with a TensorboardLogger:
        del self._logger
        self._logger = _TFKerasTensorboardLogger(
            statistics_functions=(
                statistics_functions
                if statistics_functions is not None
                else self.get_default_weight_statistics_list()
            ),
            context=context,
            tensorboard_directory=tensorboard_directory,
            run_name=run_name,
            update_frequency=update_frequency,
        )

        # Save the configurations:
        self._tracked_weights = weights

        # Initialize flags:
        self._logged_model = False
        self._logged_hyperparameters = False

    def get_weights(self) -> dict[str, Variable]:
        """
        Get the weights tensors tracked. The weights will be stored in a dictionary where each key is the weight's name
        and the value is the weight's parameter (tensor).

        :return: The weights.
        """
        return self._logger.weights

    def get_weights_statistics(self) -> dict[str, dict[str, list[float]]]:
        """
        Get the weights mean results logged. The results will be stored in a dictionary where each key is the weight's
        name and the value is a list of mean values per epoch.

        :return: The weights mean results.
        """
        return self._logger.weight_statistics

    def on_train_begin(self, logs: Optional[dict] = None):
        """
        Called once at the beginning of training process (one time call). Will log the pre-training (epoch 0)
        hyperparameters and weights.

        :param logs: Dict. Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        # The callback is on a 'fit' method - training:
        self._is_training = True

        # Start the tensorboard logger:
        self._logger.open()

        # Setup the run, logging relevant information and tracking weights:
        self._setup_run()

        # Write the initial summary of the run:
        self._logger.write_initial_summary_text()

        # Write the model's graph:
        self._logger.write_model_to_tensorboard(model=self.model)

        # Write the initial weights (epoch 0):
        self._logger.write_weights_statistics()
        self._logger.write_weights_histograms()
        self._logger.write_weights_images()

        # Make sure all values were written to the directory logs:
        self._logger.flush()

    def on_train_end(self, logs: Optional[dict] = None):
        """
        Called at the end of training, wrapping up the tensorboard logging session.

        :param logs: Currently the output of the last call to `on_epoch_end()` is passed to this argument for this
                     method but that may change in the future.
        """
        super().on_train_end()

        # Write the final run summary:
        self._logger.write_final_summary_text()

        # Close the logger:
        self._logger.close()

    def on_test_begin(self, logs: Optional[dict] = None):
        """
        Called at the beginning of evaluation or validation. Will be called on each epoch according to the validation
        per epoch configuration. In case it is an evaluation, the epoch 0 will be logged.

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        # Check if needed to mark this run as evaluation:
        if self._is_training is None:
            self._is_training = False

        # If this callback is part of evaluation and not training, need to check if the run was setup:
        if not self._is_training:
            # Start the tensorboard logger:
            self._logger.open()
            # Setup the run, logging relevant information and tracking weights:
            self._setup_run()
            # Write the initial summary of the run:
            self._logger.write_initial_summary_text()
            # Write the model's graph:
            self._logger.write_model_to_tensorboard(model=self.model)
            # Write the initial data (epoch 0):
            self._logger.write_weights_statistics()
            self._logger.write_weights_histograms()
            self._logger.write_weights_images()
            # Make sure all values were written to the directory logs:
            self._logger.flush()

    def on_test_end(self, logs: Optional[dict] = None):
        """
        Called at the end of evaluation or validation. Will be called on each epoch according to the validation
        per epoch configuration. The recent evaluation / validation results will be summarized and logged.

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        super().on_test_end(logs=logs)

        # Check if needed to end the run (in case of evaluation and not training):
        if not self._is_training:
            # Write the remaining epoch iterations results:
            self._logger.write_validation_results(ignore_update_frequency=True)
            # Write the epoch loss and metrics summaries to their graphs:
            self._logger.write_validation_summaries()
            # Write the final run summary:
            self._logger.write_final_summary_text()
            # Close the logger:
            self._logger.close()

    def on_epoch_end(self, epoch: int, logs: Optional[dict] = None):
        """
        Called at the end of an epoch, logging the current dynamic hyperparameters values, summaries and weights to
        tensorboard.

        :param epoch: Integer, index of epoch.
        :param logs:  Metric results for this training epoch, and for the validation epoch if validation is
                      performed. Validation result keys are prefixed with `val_`. For training epoch, the values of the
                      `Model`'s metrics are returned. Example : `{'loss': 0.2, 'acc': 0.7}`.
        """
        # Update the dynamic hyperparameters
        super().on_epoch_end(epoch=epoch)

        # Log the weights statistics:
        self._logger.log_weights_statistics()

        # Write the remaining epoch iterations results:
        self._logger.write_training_results(ignore_update_frequency=True)
        self._logger.write_validation_results(ignore_update_frequency=True)

        # Write the epoch text summary:
        self._logger.write_epoch_summary_text()

        # Write the epoch loss and metrics summaries to their graphs:
        self._logger.write_training_summaries()
        self._logger.write_validation_summaries()

        # Write the epoch dynamic hyperparameters values to their graphs:
        self._logger.write_dynamic_hyperparameters()

        # Write the weight histograms, images and statistics for all the tracked weights:
        self._logger.write_weights_statistics()
        self._logger.write_weights_histograms()
        self._logger.write_weights_images()

        # Make sure all values were written to the directory logs:
        self._logger.flush()

    def on_train_batch_end(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the end of a training batch in `fit` methods. The batch metrics results will be logged. If it is the
        first batch to end, the model architecture and hyperparameters will be logged as well. Note that if the
        `steps_per_execution` argument to `compile` in `tf.keras.Model` is set to `N`, this method will only be called
        every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Aggregated metric results up until this batch.
        """
        # Log the batch's results:
        super().on_train_batch_end(batch=batch, logs=logs)

        # Write the batch loss and metrics results to their graphs:
        self._logger.write_training_results()

        # Check if needed to write the hyperparameters:
        if not self._logged_hyperparameters:
            self._logger.write_parameters_table_to_tensorboard()
            self._logged_hyperparameters = True
            self._logger.write_dynamic_hyperparameters()

    def on_test_batch_end(self, batch: int, logs: Optional[dict] = None):
        """
        Called at the end of a batch in `evaluate` methods. Also called at the end of a validation batch in the `fit`
        methods, if validation data is provided. The batch metrics results will be logged. In case it is an evaluation
        run, if this was the first batch the model architecture and hyperparameters will be logged as well. Note that if
        the `steps_per_execution` argument to `compile` in `tf.keras.Model` is set to `N`, this method will only be
        called every `N` batches.

        :param batch: Integer, index of batch within the current epoch.
        :param logs:  Aggregated metric results up until this batch.
        """
        # Log the batch's results:
        super().on_test_batch_end(batch=batch, logs=logs)

        # Write the batch loss and metrics results to their graphs:
        self._logger.write_validation_results()

        # Check if needed to write the hyperparameters:
        if not self._logged_hyperparameters:
            self._logger.write_parameters_table_to_tensorboard()
            self._logged_hyperparameters = True
            self._logger.write_dynamic_hyperparameters()

    @staticmethod
    def get_default_weight_statistics_list() -> (
        list[Callable[[Union[Variable, Tensor]], Union[float, Tensor]]]
    ):
        """
        Get the default list of statistics functions being applied on the tracked weights each epoch.

        :return: The default statistics functions list.
        """
        return [tf.math.reduce_mean, tf.math.reduce_std]

    def _setup_run(self):
        """
        After the trainer / evaluator run begins, this method will be called to setup the results, hyperparameters
        and weights dictionaries for logging.
        """
        super()._setup_run()

        # Check if needed to track weights:
        if self._tracked_weights is False:
            return

        # Collect the weights for drawing histograms according to the stored configuration:
        for layer in self.model.layers:
            collect = False
            if self._tracked_weights is True:  # Collect all weights
                collect = True
            else:
                for tag in self._tracked_weights:  # Collect by given name
                    if tag in layer.name:
                        collect = True
                        break
            if collect:
                for weight_variable in layer.weights:
                    self._logger.log_weight(
                        weight_name=weight_variable.name, weight_holder=weight_variable
                    )

        # Log the initial (epoch 0) weights statistics:
        self._logger.log_weights_statistics()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing

import pydantic.v1

import mlrun.common.types

from .object import ObjectKind


class BackGroundTaskLabel(mlrun.common.types.StrEnum):
    pipeline = "pipeline"


class BackgroundTaskState(mlrun.common.types.StrEnum):
    succeeded = "succeeded"
    failed = "failed"
    running = "running"

    @staticmethod
    def terminal_states():
        return [
            BackgroundTaskState.succeeded,
            BackgroundTaskState.failed,
        ]


class BackgroundTaskMetadata(pydantic.v1.BaseModel):
    name: str
    id: typing.Optional[int]
    kind: typing.Optional[str]
    project: typing.Optional[str]
    created: typing.Optional[datetime.datetime]
    updated: typing.Optional[datetime.datetime]
    timeout: typing.Optional[int]


class BackgroundTaskSpec(pydantic.v1.BaseModel):
    pass


class BackgroundTaskStatus(pydantic.v1.BaseModel):
    state: BackgroundTaskState
    error: typing.Optional[str]


class BackgroundTask(pydantic.v1.BaseModel):
    kind: ObjectKind = pydantic.v1.Field(ObjectKind.background_task, const=True)
    metadata: BackgroundTaskMetadata
    spec: BackgroundTaskSpec
    status: BackgroundTaskStatus


class BackgroundTaskList(pydantic.v1.BaseModel):
    background_tasks: list[BackgroundTask]
